unit uHidroConEmbalseBinacional;

{$DEFINE CONCONTROLCRECIDA}
interface

uses
  Classes,
  Math, xMatDefs, uSimplex, ucosa, uGlobs, uFechas, uNodos,
  SysUtils,
  uEstados,
  uFichasLPD,
  uGeneradores,
  uconstantesSimSEE,
  uFuentesAleatorias, uCosaConNombre, uevapUruguay, math01, compol;

resourcestring
  rsGHidraulicoEmbalseBinacional = 'Generador hidráulico con embalse binacional';

const
  ro_Agua = 1000; // kg/m3 densidad del agua
  g_Gravitacion = 9.8; // N/kg constante gravitatoria

type
  THidroConEmbalseBinacional = class; // forward declaration

(* Para describir el embalse, damos el volumen máximo turbinable Vmax
que corresponde al volumen (en m3 ) que almacena el embalse entre las
cotas hmin y hmax.
Y el volumen almacenado entre hmin y hmed siendo hmed = (hmax+hmin)/2

Las cotas están medidas desde algún nivel (por ejemplo el del mar).
Para poder calcular el salto se suministra también la cota de aguas abajo
o de descarga de la turbina referida al mismo nivel. (hDescarga [m])
El nivel de descarga puede ser especificado dando su valor con la variable
hDescarga o mediante la referencia a otra central en cuyo caso se considera
que el nivel de descarga es el del lago de la otra central más el valor
hDescarga. En este caso la variable hDescarga se utiliza para tener en
cuenta cambios en la referencia de niven entre una central y otra.
Si todas las cotas están referidas al nivel del mar no hay cambios de
referencias y hDescarga se pondrá en cero.

Suponemos que el área del embalse crece linealmente con la cota
A(z) = a + b * z
V(z) = integral( A(z) )= a*z + 1/2 * b z^2

con z= (h-hmin)/(hmax-hmin)

Vmax - Vmin = V(1) = a+ 0.5 b
Vhmed - Vmin= V(0.5) = 0.5 a + 0.5^3 b

=> Vmax-Vmin - 2*(Vhmed- Vmin)= 0.5 b - 0.5^2 b =  0.5^2 b
=>  b = 4*( (Vmax-Vmin)-2* (Vhmed-Vmin) )
.:  a = Vmax-Vmin - 0.5 b

V= ( 1/2*b/dhmax^2 * dh + a/dhmax ) dh + Vmin =

Pérdida de Salto por Caudal Erogado.
====================================
El caudal erogado QErogado [m3/s] es calculado al final de cada paso
de simulación como la suma del caudal vertido más el caudal turbinado.
Dicho caudal puede afectar el nivel de aguas abajo reducciendo el salto
efectivo en la central. Para tener en cuenta este efecto se usan los
parámetros: caQE y cbQE calculados para modelar la pérdida de salto
de acuerdo a la siguiente ecuación:

dh(QE) = caQE* QE + cbQE* QE^2
*)
  TFichaHidroConEmbalseBinacional = class(TFichaLPD)
  public
    // Pagos a los generadores
    PagoPorDisponibilidad_USD_MWh: NReal; // [USD/MWh] Pago por Potencia
    PagoPorEnergia_USD_MWh: NReal; // [USD/MWh] Pago por Potencia


    hmax: NReal; //[m] cota maxima
    hmin: NReal; //[m] cota minima

    // descripción de la geometria del lago
    PuntosCotaVolumen_h: TDAOfNReal;
    PuntosCotaVolumen_V: TDAOfNReal;

    centralesAguasArriba: TListaCentralesAguasArriba {of TGeneradorHidraulico};
    hDescarga: NReal; //[m] cota de la descarga para cálculo del salto
    central_lagoDescarga: TGeneradorHidraulico;
    // Coeficientes de afectación del salto por caudal erogado
    // dh(QE) = caQE* QE + cbQE* QE^2
    caQE: NReal;
    cbQE: NReal;

    //La cota minima a partir de la cual se puede verter y la cota maxima a
    //partir de la cual el máximo vertimiento es constante
    cotaMV0, cotaMV1: NReal;
    //El caudal que se puede verter con la cota maxima.
    //QMV0 se omite porque siempre es 0
    QMV1: NReal;

    ren: NReal;       //[pu] rendimiento complexivo de turbina y generador
    Pmax_Gen: NReal;  //[MW] Potencia maxima hidraulica
    Qmax_Turb: NReal; //[m3/s]
    saltoMinimoOperativo: NReal; // [m] salto mínimo operativo de las turbinas

    fDispo: NReal;    //[pu] factor de disponibilidad fortuita
    tRepHoras: NReal;

    // filtración = filtracion_Ca + filtracion_Cb * ( h- hmin )
    filtracion_Ca: NReal; //[m3/s]
    filtracion_Cb: NReal; //[m2/s]

    // Caudal de aportes por debajo del cual consideramos ambiente MUY SECO
    // para el cálculo de la evaporación
    QaMuySeco: NReal; //[m3/s]

    HayRestriccionEmaxPasoDeTiempo: boolean; // indica si se aplica la restricción
    EmaxPasoDeTiempo: NReal; // Energía maxima generable en un paso de tiempo

    HayRestriccionQTmin: boolean; // indica si se aplica la restricción
    QTmin: NReal; // caudal mínimo para asegurar navegabilidad
    ImponerQminPorPoste: boolean;

    (* rch080402 agrego esto para dar la opción de regular el uso de Salto y Palmar *)
    flg_controlCotaObjetivoInferior: boolean;
    flg_controlCotaObjetivoSuperior: boolean;
    hObjetivo: NReal; // [m] cota objetivo
    delta_cva_ParaControlDeCota: NReal; // USD/Hm3 aplicable para control de cota
    cv_USD_Hm3_ValorizadoManual: NReal;

    // generación para Argentina en MW para cada hora del paso de tiempo
    P_AG: TDAofNReal;

    //Caudal Erogado para Cálculo de Energía Embalsada
    QE_CEE: NReal;
    hVistaMin: NReal;

    cotaControlCrecida_Inicio: NReal;
    //Altura a la cual se activa el control de crecida en m
    cotaControlCrecida_Pleno: NReal;
    //cv incurrido por Hm^3h por estar por arriba de cotaActivacionControlCrecida en USD/Hm^3h
    flg_CalcularEvaporacionDelLago: boolean;
    flg_CalcularFiltracionDelLago: boolean;


    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      hmin_, hmax_: NReal; PuntosCotaVolumen_h,
  PuntosCotaVolumen_V: TDAOfNReal; hDescarga: NReal;
  central_lagoDescarga: TGeneradorHidraulico;
  centralesAguasArriba: TListaCentralesAguasArriba; caQE, cbQE, ren_,
  Pmax_Gen_, Qmax_Turb_: NReal; fDispo_: NReal; tRepHoras: NReal;
  filtracion_Ca, filtracion_Cb, QaMuySeco: NReal; cotaMV0, cotaMV1,
  QMV1: NReal; HayRestriccionEmaxPasoDeTiempo: boolean;
  EmaxPasoDeTiempo: NReal; HayRestriccionQTmin: boolean; QTmin: NReal;
  ImponerQminPorPoste: boolean; ControlarCotaPorDebajoDeObjetivo: boolean;
  ControlarCotaPorEnciaDeObjetivo: boolean; hObjetivo: NReal;
  Delta_cvaUSD_Hm3ParaControlDeCota: NReal; cv_USD_Hm3_ValorizadoManual: NReal;
  P_AG: TDAOfNreal; QE_CEE: NReal; hVistaMin: NReal;
  saltoMinimoOperativo: NReal; cotaControlCrecida_inicio: NReal;
  cotaControlCrecida_pleno: NReal; flg_CalcularEvaporacionDelLago: boolean;
  flg_CalcularFiltracionDelLago: boolean;
  PagoPorDisponibilidad_USD_MWh: NReal; PagoPorEnergia_USD_MWh: NReal  );

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function infoAd_: string; override;
    procedure Free; override;
  end;


  THidroConEmbalseBinacional = class(TGeneradorHidraulico)
  public
    pa: TFichaHidroConEmbalseBinacional;

    fuenteDeAportes: TFuenteAleatoria;
    nombreBorne: string;
    numeroBorne: integer;

    //Variables de Estado
    NDiscH, nDiscDE: integer;
    // discretizacion de la altura y de la diferencia de enegía
    hini: NReal; // cota inicial
    dEEini: NReal; // diferencia de Energía Embalsada a favor de Uy al inicio.




    //----------- variables auxiliares
    // Calcular con cambioFicha.
    polinomioCotaVolumen: TPoliR;
    V_hmin: NReal; //[Hm3] volumen mínimo almacenable
    V_hmax: NReal; //[Hm3] volumen máximo almacenable
    V_hmed: NReal; //[Hm3] volumen con la cota media


    ro_g_ren: NReal; // ro_Agua * g_gravitacion * ren /(1E6*3600)
    dhmax: NReal; // hmax - hmin
    VTmin: NReal; // volumen turbinado mínimo en el paso para cumplir con QTmin
    QTmin_CalcVTminsobreHorasDelPaso: NReal;
    // caudal en Hm3/h para restricciones por poste.

{$IFDEF CONCONTROLCRECIDA}
    V_ControlCrecida_Inicio: NReal; // [Hm3] al que se inicia el control de crecida
{$ENDIF}

    AplicarRestriccionDeErogadoMinimo: boolean;

    h_actual: NReal; // cota actual. Es la de toma.

    // Diferencia de Energía Embalsada
    // en por unidad del máximo.
    X_dEE_actual: NReal;
    // EEmbalsadaUy-EEmbalsdaAg.  dEE(k+1) = dEE(k) + ( GenAg- GenUy )
    Xs_dEE_actual: NReal;
    // EEmbalsadaUy-EEmbalsdaAg.  dEE(k+1) = dEE(k) + ( GenAg- GenUy )

    RecorteAg: NReal; // recorte del programa Argentino en el paso de tiempo MWh
    EnergiaAG, EnergiaUY: NReal; // energías tomadas en el Paso de Tiempo MWh
    h_vistaUY, h_vistaAG: NReal;

    EnergiaAlmacenada: NReal;
    EnergiaVertida: NReal;
    EAportesPropios: NReal;


    // valor máximo de la diferencia de energía embalsada con el embalse a full.
    DEE_MAXIMA: NReal;

    // Valor máximo de la diferencia de energía embalsada determinado a partir
    // del nivel del embalse al inicio de paso.
    DEE_Max_: NReal;


    dEE_inc, dEE_dec: NReal; // entrada y salida de diferencia de energía embalsada

    // variables calculadas por paso de tiempo durante la simulación
    P_Ag_programada, P_Ag_despachada, P_AG_recortada, P_Uy_despachada: TDAOfNReal;
    // potencias despachadas para Argentina. Uruguay sale por diferencia

    salto_actual: NReal; // salto actual efectivo
{$IFDEF ITERAR}
    QErogado_IterAnterior: NReal;
    // Usamos esta variable para almacenar el QErogado en la
    // iteración anterior. Esto se hace en leerResultadosDelPaso
    // antes de cargar el nuevo valor de QErogado.
{$ENDIF}
    QMaxVert: NReal;

    ce_MAXTURB0: NReal;



    (******** Esta es la variable de estado ******)
    X_VolumenEmbalsado: NReal; // volumen actual
    Xs_VolumenEmbalsado: NReal; // volumen actual

    PmaxParaQmax: NReal; // Pontecia máxima por máquina considerando Qmax y la cota
    NMaquinasDisponibles: integer;
    Pmax_Central: TDAOfNReal; // [MW] Potencia Máxima disponible en la central por poste
    PDisponible_Central: NReal; // MW Disponibles Gen x Disp. Para remuneración.
    Pmax_Gen: NReal; // [MW] Potencia máxima generable por turbina-generador
    Qmax: NReal; // [m3/s] Caudal máximo turbinable por turbina-generador


    dh_RedQE: NReal; // [m] Guarda la pérdida de salto por caulda erogado
    // variables auxiliares para la iteración por caudal erogado
    // Si está EnIteracion, el rango de potencias se fija en
    // el procedimiento NecesitoIterar para evitar oscilaciones
    Pmax_Central_CotaInf, Pmax_Central_CotaSup: TDAOfNReal;
    // se pone a false, en sorteosDelPaso y a true al pasar por NecesitoIterar
    EnIteracion: boolean;



    PuedoGenerar: boolean;
    // indica si la Central está en condiciones de ofrecer potencia o no
    cv_agua_USD_MWh: NReal; // costo variable de generación

    cv_agua_USD_Hm3_Dec: NReal; // valor del Agua expresado en USD/MWh
    cv_dEE_USD_MWh_Dec: NReal; // valor de la diferencia de energía embalzada
    cv_dEE_USD_MWh_Inc: NReal; //valor de la diferencia de energía embalzada


    Qevap: NReal; // caudal perdido por evaporacion
    Qfilt: NReal; // caudal perdido por filtración
    VPerdidas: NReal; //( QEvap+QFilt )* dt
    // variables auxiliares para las funciones cota-volumen y su inversa
    EsConico: boolean;
    embalse_cb, embalse_ca, embalse_c1, embalse_c2, embalse_c3,
    embalse_c4, embalse_c5: NReal;

    flg_ModificandoQEmaxParaEstabilidad: boolean;
    flg_RecortandoAportesParaEstabilidad: boolean;

{$IFDEF ITERAR}
    //Coeficientes de amortiguamiento, el otro es 1 - alfa
    alfa: NReal;
{$ENDIF}

    flg_ValorizadoManual: boolean;


    // tabla para interpolación de la energía embalsada
    EAcumMax, EAcumMin, EAcumMaxMenosMin: NReal;
    hEAcumMax, hEAcumMin, dhEAcumMaxMenosMin: NReal;

    d_DEE_tablaAprox: NReal;
    kEnergiaEmbalsadaToCota: array[0..199] of NReal;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  flg_CalcularGradienteDeInversion: boolean; hIni: NReal; dEEIni: NReal;
  NDiscH, nDiscDE: integer; fuenteDeAportes: TFuenteAleatoria;
  nombreBorne: string; flg_ValorizadoManual: boolean; TonCO2xMWh: NReal;
  LowCostMustRun, CleanDevelopmentMechanism: boolean);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure RegistrarParametrosDinamicos; override;

    function PotenciaFirme: NReal; override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    //      procedure Sim_Inicio; override;
    procedure Sim_Cronica_Inicio; override;
    function get_pa_FD( kTipoUnidad: integer ): NReal; override;
    function get_pa_TMR( kTipoUnidad: integer ): NReal; override;

    procedure SorteosDelPaso(sortear: boolean); override;
    procedure PrepararPaso_ps; override;

    function ivarVertimientosMiosYAguasAbajo: TDAofNInt;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_cargue(s: TSimplex); override;
    procedure EvolucionarEstado; override;


    procedure ResolverEncadenamientos(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;

{$IFDEF ITERAR}
    function opt_necesitoIterar(kIteracion: integer;
      var errRelativo: NReal): boolean;
      override;
{$ENDIF}

{$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
{$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure PosicionarseEnEstrellita; override;
    procedure ActualizarEstadoGlobal( flg_Xs: boolean ); override;
    procedure AcumAux1(peso: NReal); override;
    procedure SetAux1; override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;
    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados);
      override;

    procedure Free; override;

    // geometría del embalse
    function CotaToEnergia(h: NReal): NReal;
    function CotaToVolumen(h: NReal): NReal;
    function VolumenToCota(Vol: NReal): NReal;
    function CotaToSuperficie(h: NReal): NReal;
    function hVerticeCono: NReal;

    procedure dEEToCota(var hvistaUy, hvistaAg: NReal; dEE_actual, h_actual: NReal);


    function CotaTodEE(h_actual, h_vista: NReal): NReal;

    procedure CambioFichaPD; override;

    procedure PubliVars; override;

    function CotaAguasArriba: NReal; override;
    // Calcula la cota de la descarga para poder calcular el Salto
    function CotaDescarga: NReal; override;
    // Caclula la reducción del salto causada por el caudal Erogado
    function ReduccionDeCotaPorCaudalErogado(QErogado: NReal): NReal; override;
    function centralLagoDescarga: TGeneradorHidraulico; override;

    // imprime tabla de valores del agua en USD/Hm3
    procedure opt_PrintResultados_Encab(var fsal: textfile); override;
    procedure opt_PrintResultados(var fsal: textfile); override;

    procedure sim_FinCronicaPrintEstadoFinal(var fsal: textfile); override;

    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;
  end;

procedure THidroConEmbalseBinacional_cambioFichaPD(Actor: TCosa);

procedure AlInicio;
procedure AlFinal;

implementation


// variable auxiliar para calculo de EnergiaToCota
var
  aux_EnergiaDada: NReal;
  aux_Central: THidroConEmbalseBinacional;


// función auxilar
function CotaToEnergia_dada(cota: NReal): NReal;
begin
  Result := aux_central.CotaToEnergia(cota) - aux_EnergiaDada;
end;

constructor THidroConEmbalseBinacional.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  flg_CalcularGradienteDeInversion: boolean;
  hIni: NReal; dEEIni: NReal; NDiscH, nDiscDE: integer;
  fuenteDeAportes: TFuenteAleatoria; nombreBorne: string;
  flg_ValorizadoManual: boolean; TonCO2xMWh: NReal;
  LowCostMustRun, CleanDevelopmentMechanism: boolean);

begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    TonCO2xMWh, LowCostMustRun, CleanDevelopmentMechanism, flg_CalcularGradienteDeInversion );
  flg_ModificandoQEmaxParaEstabilidad := False;
  flg_RecortandoAportesParaEstabilidad := False;

  Self.lpd := lpd;
  self.hini := hIni;
  self.dEEIni := dEEIni;

  self.NDiscH := NDiscH;
  self.nDiscDE := nDiscDE;
  self.fuenteDeAportes := fuenteDeAportes;
  self.nombreBorne := nombreBorne;
  self.flg_ValorizadoManual := flg_ValorizadoManual;
end;

constructor THidroConEmbalseBinacional.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);

  f.IniciarLecturaRetrasada;
  f.rd('hIni', hIni);
  f.rd('dEEIni', dEEIni);
  f.rd('NDiscH', NDiscH);
  f.rd('nDiscDE', nDiscDE);
  f.rd('lpd', TCosa(lpd));
  f.rdReferencia('fuenteDeAportes', TCosa(fuenteDeAportes), Self);
  f.rd('nombreBorne', nombreBorne);
  f.rd('flg_ValorizadoManual', flg_ValorizadoManual);
  f.EjecutarLectura;
  flg_ModificandoQEmaxParaEstabilidad := False;
  flg_RecortandoAportesParaEstabilidad := False;

  pa := nil;
  nodo := nil;
  lpd.Propietario := self;
end;

procedure THidroConEmbalseBinacional.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('hIni', hini, CF_PRECISION, CF_DECIMALES);
  f.wr('dEEIni', dEEIni);
  f.wr('NDiscH', NDiscH);
  f.wr('nDiscDE', nDiscDE);
  f.wr('lpd', lpd);
  f.wrReferencia('fuenteDeAportes', fuenteDeAportes);
  f.wr('nombreBorne', nombreBorne);
  f.wr('flg_ValorizadoManual', flg_ValorizadoManual);
end;

procedure THidroConEmbalseBinacional.prepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  setlength(P_Ag_programada, globs.NPostes);
  setlength(P_Ag_despachada, globs.NPostes);
  setlength(P_Ag_recortada, globs.NPostes);
  setlength(P_Uy_despachada, globs.NPostes);
  setlength(Pmax_Central, globs.NPostes);
  setlength(Pmax_Central_CotaInf, globs.NPostes);
  setlength(Pmax_Central_CotaSup, globs.NPostes);
  dh_RedQE := 0;
  polinomioCotaVolumen := TPoliR.Create_Init(2);
  numeroBorne := fuenteDeAportes.idBorne(nombreBorne);
end;

procedure THidroConEmbalseBinacional.registrarParametrosDinamicos;
begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA, nil,
    THidroConEmbalseBinacional_cambioFichaPD);
end;

function THidroConEmbalseBinacional.infoAd_: string;
begin
  Result := '';
end;

function THidroConEmbalseBinacional.potenciaFirme: NReal;
begin
  Result := (paUnidades.nUnidades[0]) * pa.Pmax_Gen * pa.ren;
end;

class function THidroConEmbalseBinacional.DescClase: string;
begin
  Result := rsGHidraulicoEmbalseBinacional;
end;

function THidroConEmbalseBinacional.get_pa_FD( kTipoUnidad: integer ): NReal;
begin
  result:= pa.fDispo;
end;

function THidroConEmbalseBinacional.get_pa_TMR( kTipoUnidad: integer ): NReal;
begin
  result:= pa.tRepHoras;
end;



procedure THidroConEmbalseBinacional.SorteosDelPaso(sortear: boolean);
begin

  EnIteracion := False;
  if globs.ObligarDisponibilidad_1_ then
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else if (sortear) then
  begin
    ActualizarProbabilidadesReparacionYRotura_(pa.fDispo, pa.tRepHoras);
    NMaquinasDisponibles := Sorteos_RepRotUnidades;
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    Pmax_Gen := pA.Pmax_gen * pA.fdispo;
    Qmax := pA.Qmax_Turb * pA.fdispo;
  end;
  PDisponible_Central:= NMaquinasDisponibles * Pmax_Gen;
end;

function THidroConEmbalseBinacional.ReduccionDeCotaPorCaudalErogado(
  QErogado: NReal): NReal;
begin
{$IFDEF REDUCCIONCOTAQE}
  Result := (pa.cbQE * QErogado + pa.caQE) * QErogado;
{$ELSE}
  Result := 0;
{$ENDIF}
end;

function THidroConEmbalseBinacional.centralLagoDescarga: TGeneradorHidraulico;
begin
  Result := pa.central_lagoDescarga;
end;

function THidroConEmbalseBinacional.cotaAguasArriba: NReal;
begin
  Result := h_actual;
end;

function THidroConEmbalseBinacional.CotaDescarga: NReal;
begin
  if pa.central_lagoDescarga = nil then
    Result := pa.hDescarga
  else
    Result := pa.central_LagoDescarga.cotaAguasArriba;
end;

procedure THidroConEmbalseBinacional.opt_PrintResultados_Encab(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(VolumenToCota(X_VolumenEmbalsado), ffgeneral, 6, 3));
  //  write( fsal, #9, FloatToStrF( h_vistaUY, ffgeneral, 6, 3) );
end;

procedure THidroConEmbalseBinacional.opt_PrintResultados(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(cv_agua_USD_Hm3_Dec, ffgeneral, 6, 3));
  //  write( fsal, #9, FloatToStrF( cv_dEE_USD_MWh_Dec, ffgeneral, 6,3 ) );
end;

procedure THidroConEmbalseBinacional.sim_FinCronicaPrintEstadoFinal(
  var fsal: textfile);
begin
  writeln(fsal, Nombre + '.h_actual = ' + FloatToStr(self.h_actual));
  writeln(fsal, Nombre + '.X_DEE_actual = ' + FloatToStr(self.X_DEE_actual));
end;

procedure THidroConEmbalseBinacional.dump_Variables(var f: TextFile;
  charIndentacion: char);
begin
  inherited dump_Variables(f, charIndentacion);
  writeln(f, charIndentacion, 'h_real[m]= ', FloatToStrF(h_actual, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'salto_actual[m]= ',
    FloatToStrF(salto_actual, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'QAportePropio[m^3/s]= ',
    FloatToStrF(QAportePropio, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QFilt[m^3/s]= ', FloatToStrF(QFilt, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QEvap[m^3/s]= ', FloatToStrF(QEvap, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QMaxVert[m^3/s]= ',
    FloatToStrF(QMaxVert, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'VolumenEmbalsado[hm^3]= ',
    FloatToStrF(X_VolumenEmbalsado, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VMax[hm^3]= ', FloatToStrF(V_hmax, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VAportePropio[hm^3]= ',
    FloatToStrF(VAportePropio, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VPerdidas[hm^3]= ',
    FloatToStrF(VPerdidas, ffFixed, 10, 3));
(*
  writeln(f, charIndentacion, 'cv_agua_USD_MWh_Inc[USD/MWh]= ', FloatToStrF(cv_agua_USD_MWh_Inc, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'cv_agua_USD_MWh_Dec[USD/MWh]= ', FloatToStrF(cv_agua_USD_MWh_Dec, ffFixed, 10, 3));
*)
  writeln(f);
end;

procedure THidroConEmbalseBinacional.EvolucionarEstado;
begin
  X_VolumenEmbalsado := Xs_VolumenEmbalsado;
  X_dEE_actual := Xs_dEE_actual;
  h_actual := VolumenToCota(X_VolumenEmbalsado);
  DEE_max_ := CotaToEnergia(h_actual) - EAcumMin;
  dEEToCota(h_vistaUy, h_vistaAg, X_dEE_actual, h_actual);
end;

procedure THidroConEmbalseBinacional.prepararPaso_ps;
var
  kposte, jhora, jHoraDelDia: integer;
  vaux: NReal;

  dCFdV_Inc, dCFdV_Dec: NReal;
  dCFddEEpu_Inc, dCFddEEpu_Dec: NReal;

  volMV0, volMV1: NReal;
  auxParaCalculos: NReal;

  factorRed: NReal;
  //  MargenAg: NReal;

  EEAg: NReal;
  d_RecorteAg: NReal;
  ce_MAXTURB: NReal;

{$IFDEF CONCONTROLCRECIDA}
  V_ErogadoMinimo: NReal;
  Vs_SinErogado: NReal;
  hs_SinErogado: NReal;
  Q_ErogadoMinimo: NReal;
 {$ENDIF}

begin
  {$IFDEF ITERAR}
  if globs.cntIteracionesDelPaso = 1 then
  begin
  {$ENDIF}
    if flg_ValorizadoManual then
    begin
      cv_agua_USD_Hm3_Inc := pa.cv_USD_Hm3_ValorizadoManual * globs.fActPaso;
      cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Inc;
    end
    else
    begin
    {$IFDEF DEV_CF_ESTRELLA}
      if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
      begin
        globs.CF.devxr_estrella_20_(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec);
        globs.CF.devxr_estrella_20_(ixr + 1, globs.kPaso_ + 1, dCFddEEpu_Inc,
          dCFddEEpu_Dec);
      end
      else
    {$ENDIF}
      begin
        globs.CF.devxr_continuo_20(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec);
        globs.CF.devxr_continuo_20(ixr + 1, globs.kPaso_ + 1, dCFddEEpu_Inc,
          dCFddEEpu_Dec);
      end;

      cv_agua_USD_Hm3_Inc := -dCFdV_Inc * globs.fActPaso;
      cv_agua_USD_Hm3_Dec := -dCFdV_Dec * globs.fActPaso;
      cv_dEE_USD_MWh_Inc := -dCFddEEpu_Inc * globs.fActPaso;
      cv_dEE_USD_MWh_Dec := -dCFddEEpu_Dec * globs.fActPaso;

    end;

    (**** AGREGADO PARA CONTROL DE COTA OBJETIVO DURANTE LA SIMULACION****)
    if globs.EstadoDeLaSala = CES_SIMULANDO then
    begin
      if pa.flg_controlCotaObjetivoInferior and (h_vistaUY < pa.hObjetivo) then
      begin
        cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Dec + pa.delta_cva_ParaControlDeCota;
        //    cv_agua_USD_Hm3_Inc:= cv_agua_USD_Hm3_Inc+ pa.delta_cva_ParaControlDeCota;
      end;

      if pa.flg_controlCotaObjetivoSuperior and (h_vistaUY > pa.hObjetivo) then
      begin
        cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Dec - pa.delta_cva_ParaControlDeCota;
        //    cv_agua_USD_Hm3_Inc:= cv_agua_USD_Hm3_Inc- pa.delta_cva_ParaControlDeCota;
      end;
    end;

  {$IFDEF COSTOSHIDROPOSITIVOS}
    if cv_agua_USD_Hm3_Inc < 1E-2 then
      cv_agua_USD_Hm3_Inc := 1e-2;

    if cv_agua_USD_Hm3_Dec < 1E-2 then
      cv_agua_USD_Hm3_Dec := 1e-2;
  {$ENDIF}

    dh_RedQE := ReduccionDeCotaPorCaudalErogado(NMaquinasDisponibles * Qmax);
    salto_actual := max(0.1, h_actual - CotaDescarga - dh_RedQE);
    ce_MAXTURB0 := salto_actual * ro_g_ren;
    //ro_g_ren= pA.ren * ro_Agua * g_gravitacion / (1.0E6 * 3600.0);
  {$IFDEF ITERAR}
  end;
  {$ENDIF}


  vclear(P_Ag_programada);
  for jhora := 0 to ceil( globs.HorasDelPaso - 1 ) do
  begin
    kposte := globs.kPosteHorasDelPaso[jhora];

    jHoraDelDia := globs.HoraDeInicioDelPaso + jHora;
    if (jHoraDelDia > high(pa.P_AG)) then
      jHoraDelDia := jHoraDelDia - length(pa.P_AG);

    P_Ag_programada[kposte] := P_Ag_programada[kposte] + pa.P_AG[jHoraDelDia];
  end;

  EnergiaAg := 0;
  RecorteAg := 0;
  for kposte := 0 to globs.NPostes - 1 do
  begin
    EnergiaAg := EnergiaAg + P_Ag_programada[kposte];
    P_Ag_programada[kposte] := P_Ag_programada[kposte] / globs.DurPos[kposte];
    P_Ag_despachada[kposte] := P_Ag_programada[kposte];
  end;


  self.dh_RedQE := ReduccionDeCotaPorCaudalErogado(QErogado_);
  salto_actual := max(0.1, h_actual - CotaDescarga - dh_RedQE);

  if salto_actual < pa.saltoMinimoOperativo then
    self.PuedoGenerar := False; // no puedo mover las turbinas

  ce := salto_actual * ro_g_ren;
  ce_MAXTURB := min(ce_MAXTURB0, ce);

  auxParaCalculos := 1 / (1.0E6 * ce);


  // calculamos el valor del agua en USD/MWh
  cv_agua_USD_MWh := cv_agua_USD_Hm3_Dec * auxParaCalculos;

  QAportePropio := Self.fuenteDeAportes.bornera[Self.numeroBorne];

  QAportePropio:= max(0, QAportePropio );
  (*
  Assert(QAportePropio >= 0, 'El generador hidráulico con embalse ' +
    nombre + ' tiene caudal de aportes propios negativo. Fuente: ' +
    fuenteDeAportes.nombre + ', Borne: ' + nombreBorne);
    *)

  if pa.flg_CalcularFiltracionDelLago then
    QFilt := pa.filtracion_Ca + pa.filtracion_Cb * (h_actual - pa.hmin)
  else
    QFilt := 0.0;

  if pa.flg_CalcularEvaporacionDelLago then
    QEvap := CotaToSuperficie(h_actual) * CoeficienteDeEvaporacion_mps(
      globs.MesInicioDelPaso, QAportePropio, pa.QaMuySeco)
  else
    QEvap := 0.0;


  VAportePropio := QAportePropio * globs.SegundosDelPaso / 1.0E6;
  VPerdidas := (QFilt + QEvap) * globs.SegundosDelPaso / 1.0E6;
  VPerdidas := max(0, min(VPerdidas, X_VolumenEmbalsado - V_hmin - (10 / 1.0E6)));
  // acotamos a que no pueda perderse más que lo que hay

  // en base al volumen actual calcula el vertimiento posible
  volMV0 := CotaToVolumen(pa.cotaMV0);
  if X_VolumenEmbalsado < volMV0 then
    QMaxVert := 0.1
  else
  begin
    volMV1 := CotaToVolumen(pa.cotaMV1);
    if X_VolumenEmbalsado < volMV1 then
      QMaxVert := pa.QMV1 * (X_VolumenEmbalsado - VolMV0) / (VolMV1 - VolMV0)
    else
      QMaxVert := pa.QMV1;
  end;

  PmaxParaQmax := Qmax * ce_MAXTURB * 3600;
  if NMaquinasDisponibles > 0 then
  begin
    vaux := NMaquinasDisponibles * min(Pmax_Gen, PMaxParaQmax);
    for kposte := 0 to globs.NPostes - 1 do
      Pmax_Central[kposte] := vaux;
  end
  else
    vclear(Pmax_Central);

  (**
  // si llegué al fondo para Uruguay limito la potencia
  if self.h_vistaUY <= self.pa.hVistaMin then
  begin
    vaux:= ( QAportePropio  * ce * 3600 / 2 );
   for kposte := 0 to globs.NPostes-1 do
     PMax_Central[kposte]:= min( PMax_Central[kposte], PParaArgentina[kposte]+vaux);
  end;
  **)
  EnergiaAlmacenada := cotaToEnergia(self.h_actual);


  // chequeamos que sea posible cumplir con lo programado por los hermanos argentinos.
  // si no es posible le recortamos el programa.
  for kposte := 0 to globs.NPostes - 1 do
    if P_Ag_despachada[kposte] > PMax_Central[kposte] then
    begin
      d_recorteAg := (P_Ag_despachada[kposte] - PMax_Central[kposte]) *
        globs.DurPos[kposte];
      recorteAg := recorteAg + d_recorteAg;
      energiaAg := energiaAg - d_recorteAg;
      P_Ag_despachada[kposte] := PMax_Central[kposte];
    end;


  EAportesPropios := (ce * 1.0E6) * VAportePropio; // energía de los aportes propios

  EEAg := min(EnergiaAlmacenada + EAportesPropios - X_DEE_actual -
    self.EAcumMin, // a eso tiene derecho, pero por lad dudas verifico si es posible
    EnergiaAlmacenada + EAportesPropios / 2 // esto es lo que es posible máximo
    );



  if EEAg < 0 then
    EEAg := 0;

  if (energiaAg > EEAg) then // no los dejo fondear
  begin
    factorRed := EEAg / energiaAg * 0.98;
    for kPoste := 0 to globs.NPostes - 1 do
      P_Ag_despachada[kposte] := P_Ag_despachada[kposte] * factorRed;
    EnergiaAg := EnergiaAg * factorRed;
  end;


  for kPoste := 0 to globs.NPostes - 1 do
    P_Ag_recortada[kposte] := P_Ag_programada[kposte] - P_Ag_despachada[kposte];



  PuedoGenerar := False;
  for kposte := 0 to globs.NPostes - 1 do
    if PMax_Central[kposte] > 0.1 then
    begin
      PuedoGenerar := True;
      break;
    end;

  if pA.HayRestriccionQTmin then
  begin
    VTmin := X_VolumenEmbalsado - V_hmin + VAportePropio - VPerdidas;
    VTmin := 0.95 * VTmin; // lo bajo un poco cumplo si realmente puedo
    VTmin := min(PA.QTmin * globs.SegundosDelPaso / 1.0E6, VTmin);
    QTmin_CalcVTminsobreHorasDelPaso := VTMin * globs.invHorasDelPaso;
    AplicarRestriccionDeErogadoMinimo := VTmin > 1;
  end
  else
    AplicarRestriccionDeErogadoMinimo := False;


{$IFDEF CONCONTROLCRECIDA}

  Vs_SinErogado := X_VolumenEmbalsado + VAportePropio - VPerdidas;

  // ahora chequeamos si X_VolumenEmbalsado+VAportePropio - VPerdias > V(cotamaxop)
  // si eso pasa, imponemos un ErogadoMinimo que crece linealmente
  // dese 0 a la cota del maximo operativo hasta el Máximo del vertedero
  // para una coto 2m por arriba del máximo operativo.
  // En el futuro tenemos que dejar agregar la curva de erogado mínimo


  if (Vs_SinErogado > V_ControlCrecida_Inicio) then
  begin
    hs_SinErogado := self.VolumenToCota(Vs_SinErogado);
    Q_ErogadoMinimo := pa.QMV1 * (hs_SinErogado - pa.cotaControlCrecida_Inicio) /
      (pa.cotaControlCrecida_Pleno - pa.cotaControlCrecida_Inicio);
    if QMaxVert < Q_ErogadoMinimo then
      QMaxVert := Q_ErogadoMinimo + 0.1;
    V_ErogadoMinimo := Q_ErogadoMinimo * globs.SegundosDelPaso / 1.0E6;
    if V_ErogadoMinimo > VTMin then
    begin
      //      V_ErogadoMinimo:= VTmin;
      QTmin_CalcVTminsobreHorasDelPaso := VTMin * globs.invHorasDelPaso;
      AplicarRestriccionDeErogadoMinimo := VTmin > 1;
    end;
  end;
{$ENDIF}

end;

procedure THidroConEmbalseBinacional.opt_nvers(var ivar, ivae, ires: integer);
begin
  self.ivar := ivar;

  if pa.ImponerQminPorPoste then
    ivar := ivar + globs.NPostes + globs.NPostes // PGen por poste y VVertimiento
  else
    ivar := ivar + globs.NPostes + 1; // PGen por poste y VVertimiento



{$IFDEF CONCONTROLCRECIDA}
  self.ires := ires; // 0 <= volumen_s
  ires := ires + 1;   // probando no usar <= Vmax
{$ELSE}
  // 0 <= volumen_s
  // Volumne_s <= Vmáx  ( dos restricciones )
  ires := ires + 2;
{$ENDIF}


  ires := ires + 1; //    EE - dEE > - EE_min
  // o lo que es lo mismo hvista_uy > hvista_min
  // no controlo que Argentina no vaya por abajo de h_min


  if (PuedoGenerar) and pA.HayRestriccionEmaxPasoDeTiempo then
    Inc(ires);

  if AplicarRestriccionDeErogadoMinimo then
    if pa.ImponerQminPorPoste then
      Inc(ires, globs.NPostes)
    else
      Inc(ires);
end;


function THidroConEmbalseBinacional.CotaToEnergia(h: NReal): NReal;
var
  dh: NReal;
  hD: NReal;
  c1, c2, c3: NReal;
  EnergiaEmbalsada: NReal;
begin
  dh := h - pA.hmin;
  hD := CotaDescarga + ReduccionDeCotaPorCaudalErogado(pa.QE_CEE) - pA.hmin;

  if EsConico then
  begin
    c1 := -embalse_c1 * hD;
    c2 := (embalse_c1 - 2 * embalse_c2 * hD) * 0.5;
    c3 := 2 * embalse_c2 / 3;

    EnergiaEmbalsada := ro_g_ren * ((c3 * dh + c2) * dh + c1) * dh;
  end
  else
  begin
    EnergiaEmbalsada := ro_g_ren * embalse_c1 * (0.5 * dh - hD) * dh;
  end;
  Result := EnergiaEmbalsada * 1E6;
end;



procedure THidroConEmbalseBinacional.opt_cargue(s: TSimplex);
var
  ibaseres: integer;
  iposte: integer;
  jres: integer;
  invCEPorUnoESeis: NReal;
  VolDePartida: NReal;
  aux: NReal;
  EcotaInf: NReal;

begin
{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}
  invCEPorUnoESeis := 1 / (ce * 1.0E6);
  ibaseres := nodo.ires;

  // Tenemos que aportar a las restricciones de demanda del nodo al que está
  // conectado el generador
  for iposte := 0 to globs.NPostes - 1 do
  begin
    s.pon_e(ibaseres + iposte, ivar + iposte, 1);
    s.acum_e(ibaseres + iposte, s.nc, -P_AG_despachada[iposte]);
  end;

  ivert := ivar + globs.NPostes;

  // Ahora agregamos a la función de UTILIDAD (-costo)
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(s.nf, ivar + iposte, -(cv_agua_USD_MWh + cv_dEE_USD_MWh_Dec) *
      globs.durpos[iposte]);


  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(s.nf, ivert + iposte, -cv_agua_USD_Hm3_Dec); //  Vertimiento
  end
  else
    s.pon_e(s.nf, ivert, -cv_agua_USD_Hm3_Dec); //  Vertimiento


  // no ponemos esto pue no entra en la optimización del paso.
  //  s.acum_e( s.nf, s.nc, cv_agua_USD_Hm3_Inc * VAportePropio - cv_agua_USD_Hm3_Dec * VPerdidas); // Aporte

  // restriccion de volumen >= 0 = Vmín
  jres := ires;
  if (PuedoGenerar) then
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, -globs.durpos[iposte] * invCEPorUnoESeis);
  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivert + iposte, -1);
  end
  else
    s.pon_e(jres, ivert, -1);  // vertimiento

  VolDePartida := (X_VolumenEmbalsado -V_hmin + VAportePropio - VPerdidas);
  if VolDePartida < 0.01 then
  begin
    // esto no debe pasar.
    VolDePartida := 0.01;

  end;
  s.pon_e(jres, s.nc, VolDePartida);


{$IFNDEF CONCONTROLCRECIDA}
  // restricción de volumen <= Vmax
  Inc(jres);
  if (PuedoGenerar) then
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);
  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivert + iposte, 1);
  end
  else
    s.pon_e(jres, ivert, 1);  // vertimiento
  s.pon_e(jres, s.nc, (Vmax - V_hmin) - VolDePartida);


  //??? ojo revisar si al agregar ImponerQminPorPoste no hay que tocar esto.
  //Si hay que cambiarlo, hay que permitir violar todos los postes, aunque con uno solo
  //alcanzaría
  s.permitirViolarBordeSupParaSatisfacerRestriccion(jres,
    ivarVertimientosMiosYAguasAbajo);
{$ENDIF}


(***
EReal + deeUy_s >= EAcumMin
o lo que es equivalente hVistaUy_s >= hVistaMin
************)
  Inc(jres);
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(jres, ivar + iposte, -globs.durpos[iposte]);

  // calculamos el término independiente y vemos que sea cumplible.
  aux := DEE_max_ + EAportesPropios + X_DEE_actual + 2 * EnergiaAg;



{  if (globs.kPaso_= 37)
    and (globs.CF.ordinalEstrellaActual = 1844)
    and (globs.cntIteracionesDelPaso = 8) then
    begin
      writeln('Aplicando ajuste Aux:', aux, ' EnergiaAg: ', energiaAg );
      writeln('opatropa');
    end;}

  if (aux < (EnergiaAg + 10)) then
  begin
    // si es necesario aflojo para poder cumplir con el programa Argentino
    aux := EnergiaAg + 10;
    // podríamos marcar la restricción como de igualdad
    // restauro el límite inferior para que sea factible
    for iposte := 0 to globs.NPostes - 1 do
      Pmax_Central_CotaInf[iposte] := P_Ag_despachada[iposte];
  end
  else
  begin
    // si estoy iterando controlo que no halla convertido en infactible el problema
    // al mover las cotas inferiores de la potencia para la convergencia de la iteración
    // si es así las aflojo
    ECotaInf := 0;
    for iposte := 0 to globs.NPostes - 1 do
    begin
      ECotaInf := ECotaInf + PMax_Central_CotaInf[iposte] * globs.DurPos[iposte];
    end;
    if (aux < (ECotaInf + 10)) then
    begin
      // Volvemos a aflojar las cotas inferiores de las potencias
      for iposte := 0 to globs.NPostes - 1 do
        Pmax_Central_CotaInf[iposte] := P_Ag_despachada[iposte];
    end;
  end;



  s.pon_e(jres, s.nc, aux);



(*************
Ereal - deeUy_s >= EAcumMin
o lo que es equivalente hVistaAg_s >= hVistaMin
*************)
(********** no pongo esta
  inc( jres );
  for iposte:= 0 to globs.NPostes - 1 do
    s.pon_e( jres, ivar+iposte, +globs.durpos[iposte] );
  s.pon_e(
    jres, s.nc,
    dEE_Max - 2*EnergiaAg - deeUy -EAportesPropios);
****)


  // Restricción de Energía máxima en el paso de tiempo
  if (PuedoGenerar) and pA.HayRestriccionEmaxPasoDeTiempo then
  begin
    Inc(jres);
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, -globs.durpos[iposte]);
    s.pon_e(jres, s.nc, pA.EmaxPasoDeTiempo);
  end;

  // Restricción de caudal erogado mínimo por navegabilidad
  if AplicarRestriccionDeErogadoMinimo then
  begin
    Inc(jres);
    if pa.ImponerQminPorPoste then
    begin
      for iposte := 0 to globs.NPostes - 1 do
      begin
        if (PuedoGenerar) then
        begin
          s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);
        end;
        s.pon_e(jres, ivert + iposte, 1);  // vertimiento
        s.pon_e(jres, s.nc, -QTmin_CalcVTminsobreHorasDelPaso * globs.DurPos[iposte]);
        Inc(jres);
      end;
    end
    else
    begin
      if (PuedoGenerar) then
      begin
        for iposte := 0 to globs.NPostes - 1 do
          s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);
      end;
      s.pon_e(jres, ivert, 1);  // vertimiento
      s.pon_e(jres, s.nc, -VTmin);
    end;
  end;
end;




procedure THidroConEmbalseBinacional.ResolverEncadenamientos(s: TSimplex);
var
  iCentrales: integer;
  gh: TFichaCentralAguasArriba;
begin
  // Aportes de las centrales aguas arriba
  // sumamos los aportes de las centrales aguas arriba
  // sobre la restricción de vertimiento
  for iCentrales := 0 to pa.centralesAguasArriba.Count - 1 do
  begin
    gh := TFichaCentralAguasArriba(pa.centralesAguasArriba[iCentrales]);
    gh.Central.CargarAportesAFilaCentralAguasAbajo(
      ires, // fila en la que cargar los aportes
      gh.coef, // coeficiente de llegada por el que multiplicar los aportes
      self.cv_agua_USD_Hm3_Inc,
      // valor del agua a considerar en la central aguas abajo.
      s);
  end;
{$IFNDEF CONCONTROLCRECIDA}
  for iCentrales := 0 to pa.centralesAguasArriba.Count - 1 do
  begin
    gh.Central.CargarAportesAFilaCentralAguasAbajo(
      ires + 1, // fila en la que cargar los aportes
      -gh.coef, // coeficiente de llegada por el que multiplicar los aportes
      0,   // si había algo que cargar lo cargó la anterior.
      s);
  end;
{$ENDIF}
end;


procedure THidroConEmbalseBinacional.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iposte: integer;
  qmax_Hm3PorHora: double;
begin
  // Le fijamos como cota máxima PMax a la potencia en todos los postes
  if PuedoGenerar then
  begin

    if EnIteracion then
    begin
      for iposte := 0 to globs.NPostes - 1 do
        s.cota_inf_set(ivar + iposte, PMax_Central_CotaInf[iposte]);

      for iposte := 0 to globs.NPostes - 1 do
        s.cota_sup_set(ivar + iposte, PMax_Central_CotaSup[iposte]);
    end
    else
      for iposte := 0 to globs.NPostes - 1 do
      begin
        s.cota_inf_set(ivar + iposte, P_Ag_despachada[iposte]);
        s.cota_sup_set(ivar + iposte, PMax_Central[iposte]);
      end;

  end
  else
    for iposte := 0 to globs.NPostes - 1 do
      s.FijarVariable(ivar + iposte, 0);


  if pa.ImponerQminPorPoste then
  begin
    qmax_Hm3PorHora := QMaxVert * globs.SegundosDelPaso / 1.0E6 / globs.HorasDelPaso;
    for iposte := 0 to globs.NPostes - 1 do
      s.cota_sup_set(ivert + iposte, qmax_Hm3PorHora * globs.DurPos[iposte]);
  end
  else
    s.cota_sup_set(ivert, QMaxVert * globs.SegundosDelPaso / 1.0E6);
end;

procedure THidroConEmbalseBinacional.dEEToCota(var hvistaUy, hvistaAg: NReal;
  dEE_actual, h_actual: NReal);

var
  EEReal: NReal;

  function EEToCota(EE: NReal): NReal;
  var
    res: NReal;
    kEE: integer;
  begin
    kEE := trunc((EE - EAcumMin) / EAcumMaxMenosMin * 199 + 0.5);
    if (kEE > 199) then
    begin
      res := kEnergiaEmbalsadaToCota[199] + (kEnergiaEmbalsadaToCota[199] -
        kEnergiaEmbalsadaToCota[198]) * (kEE - 199);
    end
    else
    if kEE >= 0 then
      res := kEnergiaEmbalsadaToCota[kEE]
    else
    begin
      res := kEnergiaEmbalsadaToCota[0] + (kEnergiaEmbalsadaToCota[1] -
        kEnergiaEmbalsadaToCota[0]) * kEE;
    end;
    Result := res;
  end;

begin
  EEReal := CotaToEnergia(h_actual);

  if abs(dEE_actual) > d_DEE_tablaAprox then
  begin
    hVistaUy := EEToCota(EEREal + dEE_actual);
    hVistaAg := EEToCota(EEReal - dEE_actual);
    assert((hVistaUy - h_actual) * (hVistaAg - h_actual) <= 0,
      'Error,THidroConEmbalseBinacional.dEEToCota. Cotas vistas del mismo lado.');
  end
  else
  begin
    hVistaUy := h_actual;
    hVistaAg := h_actual;
  end;
end;

function THidroConEmbalseBinacional.CotaTodEE(h_actual, h_vista: NReal): NReal;
var
  EEReal, EEUy: NReal;
  dEE_actual: NReal;
begin
  EEReal := CotaToEnergia(h_actual);
  EEUy := CotaToEnergia(h_vista);

  dEE_actual := EEUy - EEReal;

  Result := dEE_actual;
end;

procedure THidroConEmbalseBinacional.Sim_Cronica_Inicio;
begin
  inherited Sim_Cronica_Inicio;
  QErogado_ := 0;
{$IFDEF ITERAR}
  QErogado_IterAnterior := 0; // para que de igual al correrlo dos veces seguidas
{$ENDIF}
  VErogado := 0;
  QVertido := 0;
  VVertido := 0;
  QTurbinado := 0;
  VTurbinado := 0;
  EnergiaGenerada := 0;
  X_VolumenEmbalsado := CotaToVolumen(hini);

  h_actual := hini;

  // suponemos aportes cero, en preparar paso_ps que haga las cuentas bien.
  DEE_max_ := CotaToEnergia(h_actual) - EAcumMin;
  X_DEE_actual := DEEini;

  dEEToCota(h_vistaUy, h_vistaAg, X_DEE_actual, h_actual);
end;

procedure THidroConEmbalseBinacional.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
begin
  if PuedoGenerar then
  begin
    // recuperamos los valores de Potencia despachada
    EnergiaGenerada := 0;
    for iposte := 0 to globs.NPostes - 1 do
    begin
      P[iposte] := s.xval(ivar + iposte);
      P_Uy_despachada[iposte] := P[iposte] - P_Ag_despachada[iposte];
      EnergiaGenerada := EnergiaGenerada + P[iposte] * globs.durpos[iposte];
    end;
    EnergiaUy := EnergiaGenerada - EnergiaAG;
    Ingreso_PorEnergia:= EnergiaUy * pa.PagoPorEnergia_USD_MWh;
  end
  else
  begin
    for iposte := 0 to globs.NPostes - 1 do
    begin
      P[iposte] := 0;
      P_Ag_recortada[iposte] := P_Ag_programada[iposte];
    end;
    EnergiaGenerada := 0;
    EnergiaAg := 0;
    EnergiaUy := 0;
    Ingreso_PorEnergia:= 0;
  end;


  Ingreso_PorDisponibilidad:= PDisponible_Central * pa.PagoPorDisponibilidad_USD_MWh;
  costoDirectoDelPaso:= Ingreso_PorEnergia + Ingreso_PorDisponibilidad;

  if pa.ImponerQminPorPoste then
  begin
    VVertido := 0;

    for iposte := 0 to globs.NPostes - 1 do
      VVertido := VVertido + s.xval(ivert + iposte);
  end
  else
    VVertido := s.xval(ivert);


  VTurbinado := EnergiaGenerada / (ce * 1.0E6);
  VErogado := VTurbinado + VVertido;

  QVertido := VVertido * (1.0E6 * globs.invSegundosDelPaso);
  QTurbinado := VTurbinado * (1.0E6 * globs.invSegundosDelPaso);
  QErogado_ := VErogado * (1.0E6 * globs.invSegundosDelPaso);


  Xs_VolumenEmbalsado := s.yval(ires) + V_hmin;

  dEE_Dec := EnergiaUy;
  dEE_Inc := EnergiaAg;


  Xs_DEE_actual := X_DEE_actual + (dEE_Inc - dEE_Dec);

  if (VVertido > 0.1) then
  begin
    EnergiaVertida := VVertido * (ce * 1.0E6);
    if Xs_dEE_actual > 0 then
    begin
      if Xs_dEE_actual > EnergiaVertida then
      begin
        dEE_Dec := dEE_Dec + EnergiaVertida;
        Xs_dEE_actual := Xs_dEE_actual - EnergiaVertida;
      end
      else
      begin
        dEE_Dec := dEE_Dec + Xs_dEE_actual;
        Xs_dEE_actual := 0;
      end;
    end
    else
    begin
      if (-Xs_dEE_actual) > EnergiaVertida then
      begin
        dEE_Inc := dEE_Inc + EnergiaVertida;
        Xs_dEE_actual := Xs_dEE_actual + EnergiaVertida;
      end
      else
      begin
        dEE_Inc := dEE_Inc - Xs_dEE_actual;
        Xs_dEE_actual := 0;
      end;
    end;
  end;

end;

{$IFDEF ITERAR}
function THidroConEmbalseBinacional.opt_NecesitoIterar(kIteracion: integer;
  var errRelativo: NReal): boolean;
var
  qm: NReal;
  alfa: NReal;
  k: integer;
begin
  qm := max(QErogado_, QErogado_IterAnterior);
  if qm > 1 then
    errRelativo := abs(QErogado_ - QErogado_IterAnterior) / qm
  else
    errRelativo := 0;

  if not EnIteracion then
  begin
    for k := 0 to globs.NPostes - 1 do
    begin
      Pmax_Central_CotaInf[k] := P_Ag_despachada[k];
      Pmax_Central_CotaSup[k] := Pmax_Central[k];
    end;
    EnIteracion := True;
  end;

   (****************???????????????????
    if Pmax_Central_CotaInf[k] < P_Ag_despachada[k] then
       Pmax_Central_CotaInf[k] :=  P_Ag_despachada[k];
       ******************)



  for k := 0 to globs.NPostes - 1 do
  begin
    Pmax_Central_CotaInf[k] := (0.9 * Pmax_Central_CotaInf[k] + 0.1 * P[k]);
    Pmax_Central_CotaSup[k] :=
      min((Pmax_Central_CotaSup[k] + P[k]) / 2.0, PMax_Central[k]);
  end;

  alfa := 0.7;
  QErogado_ := alfa * QErogado_IterAnterior + (1 - alfa) * QErogado_;
  QErogado_IterAnterior := QErogado_;
  Result := errRelativo > 0.1;
end;

{$ENDIF}

{$IFDEF SPXCONLOG}
procedure THidroConEmbalseBinacional.spx_NombrarVariables(s: TSimplex);
var
  iposte, jres: integer;
begin
  for iposte := 0 to globs.NPostes - 1 do
    s.set_NombreVar(ivar + iposte, Nombre + '_P[MW]' + IntToStr(iposte + 1));

  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.set_NombreVar(ivar + globs.NPostes + iposte,
        Nombre + '_Vertimiento[Hm3]' + IntToStr(iposte + 1));
  end
  else
    s.set_NombreVar(ivar + globs.NPostes, Nombre + '_Vertimiento[Hm3]');

  jres := ires;
  s.set_NombreRest(jres, Nombre + '_VOL>=0');

  {$IFNDEF CONCONTROLCRECIDA}
  Inc(jres);
  s.set_NombreRest(jres, Nombre + '_VOL<=Vmax');
  {$ENDIF}

  Inc(jres);
  if PuedoGenerar and pA.HayRestriccionEmaxPasoDeTiempo then
  begin
    s.set_NombreRest(jres, Nombre + '_Emax');
    Inc(jres);
  end;
  if AplicarRestriccionDeErogadoMinimo then
    if pa.ImponerQminPorPoste then
    begin
      for iposte := 0 to globs.NPostes - 1 do
        s.set_NombreRest(jres + iposte, Nombre + '_QTmin' + IntToStr(iposte + 1));
    end
    else
      s.set_NombreRest(jres, Nombre + '_QTmin');

end;

{$ENDIF}

function THidroConEmbalseBinacional.getNombreVar(ivar: integer;
  var nombre: string): boolean;
var
  iaux: integer;
begin
  if (ivar >= self.ivar) and (ivar < self.ivar + globs.NPostes) then
  begin
    nombre := self.Nombre + '_P[MW]' + IntToStr(ivar - self.ivar + 1);
    Result := True;
  end
  else

  if pa.ImponerQminPorPoste then
  begin
    iaux := ivar - ivert + 1;
    if (iaux >= 1) and (iaux <= globs.npostes) then
    begin
      nombre := self.Nombre + '_Vertimiento[Hm3]' + IntToStr(iaux);
      Result := True;
    end
    else
      Result := False;
  end
  else
  begin
    if (ivar = ivert) then
    begin
      nombre := self.Nombre + '_Vertimiento[Hm3]';
      Result := True;
    end
    else
      Result := False;
  end;
end;

function THidroConEmbalseBinacional.getNombreRes(ires: integer;
  var nombre: string): boolean;
var
  kres, cnt_res: integer;
  buscando: boolean;
  res: boolean;

  function check(condicion: boolean; cnt: integer; texto: string): boolean;
  var
    icond: integer;
    res: boolean;
  begin
    res := False;
    if buscando and condicion then
    begin
      for icond := 0 to cnt - 1 do
      begin
        if kres = cnt_res then
        begin
          nombre := self.nombre + texto + '_' + IntToStr(icond + 1);
          buscando := False;
          res := True;
          break;
        end
        else
          Inc(cnt_res);
      end;
    end;
    Result := res;
  end;

begin
  kres := ires - self.ires + 1;
  cnt_res := 1;
  buscando := True;
  res :=
    check(True, 1, '_VOL>=0') or check(True, 1, '_VOL>=0');

  {$IFNDEF CONCONTROLCRECIDA}
  res := res or check(True, 1, '_VOL<=Vmax');
  {$ENDIF}

  if (PuedoGenerar) then
    res := res or check(True, 1, 'hvista_uy > hv_min') or
      check(pa.HayRestriccionEmaxPasoDeTiempo, 1, '_res-EMax');

  if pa.ImponerQminPorPoste then
    res := res or check(AplicarRestriccionDeErogadoMinimo, globs.NPostes,
      '_res-QTMin')
  else
    res := res or check(AplicarRestriccionDeErogadoMinimo, 1, '_res-QTMin');

  Result := res;
end;

procedure THidroConEmbalseBinacional.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableVR('P_Uy_despachada', '[MW]', 6, 1, P_Uy_despachada, True, True);
  PublicarVariableVR('P_Ag_despachada', '[MW]', 6, 1, P_Ag_despachada, True, True);
  PublicarVariableVR('P_Ag_recortada', '[MW]', 6, 1, P_Ag_recortada, True, True);

  PublicarVariableNR('QAportesP', '[m3/s]', 8, 1, QAportePropio, True);
  PublicarVariableNR('QTurbinado', '[m3/s]', 8, 1, QTurbinado, True);
  PublicarVariableNR('QVertido', '[m3/s]', 8, 1, QVertido, True);
  PublicarVariableNR('h_real', '[m]', 6, 1, h_actual, True);
  PublicarVariableNR('h_vistaUY', '[m]', 6, 1, h_vistaUY, True);
  PublicarVariableNR('h_vistaAG', '[m]', 6, 1, h_vistaAG, True);
  PublicarVariableNR('dh_RedQE', '[USD/Hm3]', 6, 2, dh_RedQE, True);
  PublicarVariableNR('CV_aguaDec', '[USD/Hm3]', 6, 1, cv_agua_USD_Hm3_Dec, True);
  PublicarVariableNR('CV_aguaInc', '[USD/Hm3]', 6, 1, cv_agua_USD_Hm3_Inc, True);
  PublicarVariableNR('CV_DEE_Dec', '[USD/MWh]', 6, 1, cv_dEE_USD_MWh_Dec, True);
  PublicarVariableNR('DEE', '[MWh]', 6, 1, X_DEE_actual, True);

  PublicarVariableVR('P', '[MW]', 6, 1, P, True, False);
  PublicarVariableNR('CV_DEE_Inc', '[USD/MWh]', 6, 1, cv_dEE_USD_MWh_Inc, False);
  PublicarVariableNR('Vertimiento', '[Hm^3]', 15, 1, VVertido, False);
  PublicarVariableNR('Cota', '[m]', 6, 1, h_actual, False);
  PublicarVariableNR('Vol', '[Hm^3]', 15, 1, X_VolumenEmbalsado, False);
  PublicarVariableNI('NMaqsDisponibles', '-', NMaquinasDisponibles, False);
  {$IFDEF ITERAR}
  PublicarVariableNR('QErogado_IterAnterior', '[m3/s]', 8, 1,
    QErogado_IterAnterior, False);
  {$ENDIF}
  PublicarVariableNR('QErogado', '[m3/s]', 8, 1, QErogado_, False);
  PublicarVariableNR('QEvaporación', '[m3/s]', 8, 1, QEvap, False);
  PublicarVariableNR('QFiltración', '[m3/s]', 8, 1, QFilt, False);
end;

function THidroConEmbalseBinacional.ivarVertimientosMiosYAguasAbajo: TDAofNInt;
var
  nCentrales: integer;
  iter: TGeneradorHidraulico;
  res: TDAofNInt;
begin
  nCentrales := 1; //yo

  iter := centralLagoDescarga;  // agregado
  while iter <> nil do
  begin
    nCentrales := nCentrales + 1;
    iter := iter.centralLagoDescarga;
  end;

  SetLength(res, nCentrales);
  nCentrales := 0;
  iter := self;
  while iter <> nil do
  begin
    res[nCentrales] := iter.ivert;
    nCentrales := nCentrales + 1;
    iter := iter.centralLagoDescarga;
  end;
  Result := res;
end;

procedure THidroConEmbalseBinacional.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  if not flg_ValorizadoManual then
  begin
    self.ixr := ixr;
    ixr := ixr + 2;

    Self.iauxr := iauxr;
    iauxr := iauxr + 1;
  end;
end;

procedure THidroConEmbalseBinacional.PosicionarseEnEstrellita;
begin
  EnIteracion := False;
  if not flg_ValorizadoManual then
  begin
    X_VolumenEmbalsado := globs.CF.xr[ixr];
    h_actual := VolumenToCota(X_VolumenEmbalsado);
    QErogado_ := globs.Auxs_r0[iauxr][globs.CF.ordinalEstrellaActual];
{$IFDEF ITERAR}
    QErogado_IterAnterior := QErogado_;
{$ENDIF}
    X_DEE_actual := globs.CF.xr[ixr + 1];
    DEE_max_ := CotaToEnergia(h_actual) - EAcumMin;
    DEEToCota(h_vistaUy, h_vistaAg, X_DEE_actual, h_actual);
  end;
end;

procedure THidroConEmbalseBinacional.ActualizarEstadoGlobal( flg_Xs: boolean );
begin
  if flg_ValorizadoManual then
    exit;
  globs.CF.xr[ixr] := X_VolumenEmbalsado;
  globs.CF.xr[ixr + 1] := X_DEE_actual;
end;

procedure THidroConEmbalseBinacional.AcumAux1(peso: NReal);
begin
  if flg_ValorizadoManual then
    exit;
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] :=
    globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] + Self.QErogado_ * peso;
end;

procedure THidroConEmbalseBinacional.SetAux1;
begin
  if flg_ValorizadoManual then
    exit;
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] := Self.QErogado_;
end;

procedure THidroConEmbalseBinacional.optx_RegistrarVariablesDeEstado(
  adminEstados: TAdminEstados);
begin
  if flg_ValorizadoManual then
    exit;

  adminEstados.Registrar_Continua(
    ixr,
    V_hmin,
    V_hmax,
    NDiscH,
    nombre + '_Vol', // nombre de la variable
    'Hm3' // unidades
    );

  adminEstados.RegistrarTraduccion_Continua(ixr,
    nombre + '_Cota',
    'm', self.VolumenToCota);



  adminEstados.Registrar_Continua(
    ixr + 1, -DEE_MAXIMA,
    DEE_MAXIMA,
    self.nDiscDE,
    nombre + '_dEE', // nombre de la variable
    'MWh' // unidades
    );
end;

procedure THidroConEmbalseBinacional.Free;
begin
  lpd.Free;
  setlength(P_Uy_despachada, 0);
  setlength(P_Ag_programada, 0);
  setlength(P_Ag_despachada, 0);
  setlength(P_Ag_recortada, 0);
  setlength(Pmax_Central, 0);
  setlength(Pmax_Central_CotaInf, 0);
  setlength(Pmax_Central_CotaSup, 0);

  if polinomioCotaVolumen <> nil then
    polinomioCotaVolumen.Free;
  inherited Free;

end;




function THidroConEmbalseBinacional.CotaToVolumen(h: NReal): NReal;
var
  dh: NReal;
begin
  dh := h - pA.hmin;
  if EsConico then
    Result := (embalse_c2 * dh + embalse_c1) * dh
  else
    Result := embalse_c1 * dh;
end;



function THidroConEmbalseBinacional.CotaToSuperficie(h: NReal): NReal;
var
  dh: NReal;
begin
  dh := h - pA.hmin;
  if EsConico then
    Result := (2 * embalse_c2 * dh + embalse_c1) * 1.0E6
  else
    Result := embalse_c1 * 1.0E6;
end;


function THidroConEmbalseBinacional.hVerticeCono: NReal;
begin
  Result := pA.hmin - embalse_c1 / (2 * embalse_c2);
end;


function THidroConEmbalseBinacional.VolumenToCota(Vol: NReal): NReal;
begin
  if EsConico then
    Result := sqrt(embalse_c3 * Vol + embalse_c4) + embalse_c5
  else
    Result := Vol / embalse_c1 + pA.hmin;
end;

procedure THidroConEmbalseBinacional.cambioFichaPD;
var
  //  h: NReal;
  dh: NReal;
  //  QE: NReal;
  j: integer;
  Eaux: NReal;
  //  salto: NReal;
  //  perdida_hPorQE: NReal;
  //  buscando: boolean;
  h1{, h2, E1, E2}: NReal;

  //  dEEMax: NReal;
  EacumEmbalse: NReal;
  Eacum: NReal;
  MaximoDeLaEnergiaEmbalsada: NReal;
  //  hVertice: NReal;

  cntIters: word;
  converged: boolean;

  hmed: NReal;

  dVmax, dVmed: NReal;
begin

  polinomioCotaVolumen.AproxTabla2D(2, self.pa.PuntosCotaVolumen_h,
    self.pa.PuntosCotaVolumen_V);
  V_hmin := polinomioCotaVolumen.valx(pa.hmin);
  V_hmax := polinomioCotaVolumen.valx(pA.hmax);
  hmed := (pa.hmax + pa.hmin) / 2;
  V_hmed := polinomioCotaVolumen.valx(hmed);

  dVmax:= V_hmax - V_hmin;
  dVmed:= V_hmed - V_hmin;


  //TODO maxDiferenciaEnergiaEmbalsada
  //  dEEMax:= pa.maxDiferenciaEnergiaEmbalsada;

  ro_g_ren := pA.ren * ro_Agua * g_gravitacion / (1.0E6 * 3600.0);

  dhmax := (pA.hmax - pA.hmin);

  embalse_cb := 4.0 * (dVmax - 2.0 * dVmed);
  embalse_ca := dVmax - embalse_cb / 2.0;
  if (embalse_cb < 0) then
    raise Exception.Create('Revise los datos del embalse(' + Nombre +
      '), el área decrece al aumentar la cota.');
  if (embalse_ca < 0) then
    raise Exception.Create('Revise los datos del embalse(' + Nombre +
      '), el volumen parece anularse antes de la cota mínima.');

  embalse_c1 := embalse_ca / dhmax;

  if EsCero(embalse_cb) then
  begin
    EsConico := False;
  end
  else
  begin
    EsConico := True;
    embalse_c2 := embalse_cb / 2.0 / sqr(dhmax);
    embalse_c3 := 1 / embalse_c2;
    embalse_c4 := sqr(embalse_c1 / (2 * embalse_c2));
    embalse_c5 := pA.hmin - (embalse_c1 / (2 * embalse_c2));
  end;


  {$IFDEF CONCONTROLCRECIDA}
    self.V_ControlCrecida_Inicio := self.CotaToVolumen(pa.cotaControlCrecida_Inicio);
  {$ENDIF}


  // recalculo de la energía embalsada.
  //  QE:= pa.QE_CEE;

  dh := (pA.hmax - pA.hmin) / 199;

  EAcumEmbalse := CotaToEnergia(pa.hmax);

  //  hVertice:= hVerticeCono;

  hEAcumMin := pa.hVistaMin;
  eAcumMin := cotaToEnergia(hEAcumMin);

  DEE_MAXIMA := (EAcumEmbalse - eAcumMin);

  // buscamos hEAcumMax
  EAcumMax := EAcumEmbalse + DEE_MAXIMA;

  aux_Central := self;
  aux_EnergiaDada := EAcumMax;
  hEAcumMax := pa.hmax + (pa.hmax - hEAcumMin);

  Dicot(
    CotaToEnergia_dada,               {funci¢n a anular}
    hEAcumMin, hEAcumMax, dh,          {extremos y tolerancia}
    100,          {n£mero m ximo de iteraciones}
    hEAcumMax, EAcum,     {ra¡z y f(ra¡z)}
    cntIters,          {n£mero de iteraciones realizadas}
    converged);    {valid‚s del resultado}

  // calculo de la inversa
  MaximoDeLaEnergiaEmbalsada := eAcumMax;

  d_DEE_TablaAprox := (MaximoDeLaEnergiaEmbalsada - eAcumMin) / 199;
  kEnergiaEmbalsadaToCota[0] := hEAcumMin;
  kEnergiaEmbalsadaToCota[199] := hEAcumMax;
  h1 := hEAcumMin;



  for j := 1 to 198 do
  begin
    Eaux := d_DEE_TablaAprox * j + EAcumMin;
    aux_EnergiaDada := EAux;
    Dicot(
      CotaToEnergia_Dada,               {funci¢n a anular}
      h1, hEAcumMax, dh,          {extremos y tolerancia}
      100,          {n£mero m ximo de iteraciones}
      h1, EAcum,     {ra¡z y f(ra¡z)}
      cntIters,          {n£mero de iteraciones realizadas}
      converged);    {valid‚s del resultado}

    kEnergiaEmbalsadaToCota[j] := h1;

  end;


  EAcumMaxMenosMin := EAcumMax - EAcumMin;
  dhEAcumMaxMenosMin := hEAcumMax - hEAcumMin;
end;

{*********************************
*Métodos de TFichaHidroConEmbalseBinacional*
*********************************}

constructor TFichaHidroConEmbalseBinacional.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; hmin_, hmax_: NReal;
  PuntosCotaVolumen_h, PuntosCotaVolumen_V: TDAOfNReal; hDescarga: NReal;
  central_lagoDescarga: TGeneradorHidraulico;
  centralesAguasArriba: TListaCentralesAguasArriba;
  caQE, cbQE, ren_, Pmax_Gen_, Qmax_Turb_: NReal; fDispo_: NReal;
  tRepHoras: NReal; filtracion_Ca, filtracion_Cb, QaMuySeco: NReal;
  cotaMV0, cotaMV1, QMV1: NReal; HayRestriccionEmaxPasoDeTiempo: boolean;
  EmaxPasoDeTiempo: NReal; HayRestriccionQTmin: boolean; QTmin: NReal;
  ImponerQminPorPoste: boolean; ControlarCotaPorDebajoDeObjetivo: boolean;
  ControlarCotaPorEnciaDeObjetivo: boolean; hObjetivo: NReal;
  Delta_cvaUSD_Hm3ParaControlDeCota: NReal; cv_USD_Hm3_ValorizadoManual: NReal;
  P_AG: TDAOfNreal; QE_CEE: NReal; hVistaMin: NReal; saltoMinimoOperativo: NReal;
  cotaControlCrecida_inicio: NReal; cotaControlCrecida_pleno: NReal;
  flg_CalcularEvaporacionDelLago: boolean; flg_CalcularFiltracionDelLago: boolean;
  PagoPorDisponibilidad_USD_MWh: NReal; PagoPorEnergia_USD_MWh: NReal  );
begin
  inherited Create(capa, fecha, periodicidad);
  self.hmin := hmin_;
  self.hmax := hmax_;

  self.PuntosCotaVolumen_h := PuntosCotaVolumen_h;
  self.PuntosCotaVolumen_V := PuntosCotaVolumen_V;

  self.hDescarga := hDescarga;
  self.central_lagoDescarga := central_lagoDescarga;
  self.caQE := caQE;
  self.cbQE := cbQE;
  self.ren := ren_;
  self.Pmax_Gen := Pmax_Gen_;
  self.Qmax_Turb := Qmax_Turb_;
  self.fDispo := fDispo_;
  self.tRepHoras := tRepHoras;

  Self.filtracion_Ca := filtracion_Ca;
  Self.filtracion_Cb := filtracion_Cb;
  Self.QaMuySeco := QaMuySeco;

  self.cotaMV0 := cotaMV0;
  self.cotaMV1 := cotaMV1;
  self.QMV1 := QMV1;

  self.centralesAguasArriba := centralesAguasArriba;

  self.HayRestriccionEmaxPasoDeTiempo := HayRestriccionEmaxPasoDeTiempo;
  self.EmaxPasoDeTiempo := EmaxPasoDeTiempo;
  self.HayRestriccionQTmin := HayRestriccionQTmin;
  self.QTmin := QTmin;
  self.ImponerQminPorPoste := ImponerQminPorPoste;

  self.flg_controlCotaObjetivoInferior := ControlarCotaPorDebajoDeObjetivo;
  self.flg_controlCotaObjetivoSuperior := ControlarCotaPorEnciaDeObjetivo;
  self.hObjetivo := hObjetivo;
  self.delta_cva_ParaControlDeCota := Delta_cvaUSD_Hm3ParaControlDeCota;
  self.cv_USD_Hm3_ValorizadoManual := cv_USD_Hm3_ValorizadoManual;

  self.P_AG := P_AG;

  self.QE_CEE := QE_CEE;
  self.hVistaMin := hVistaMin;
  self.saltoMinimoOperativo := saltoMinimoOperativo;
  self.cotaControlCrecida_inicio := cotaControlCrecida_inicio;
  self.cotaControlCrecida_pleno := cotaControlCrecida_pleno;

  self.flg_CalcularEvaporacionDelLago := flg_CalcularEvaporacionDelLago;
  self.flg_CalcularFiltracionDelLago := flg_CalcularFiltracionDelLago;
  self.PagoPorDisponibilidad_USD_MWh := PagoPorDisponibilidad_USD_MWh;
  self.PagoPorEnergia_USD_MWh := PagoPorEnergia_USD_MWh;

end;

procedure TFichaHidroConEmbalseBinacional.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);

  f.wr('hmin', hmin, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('hmax', hmax, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('PuntosCotaVolumen_h', PuntosCotaVolumen_h, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('PuntosCotaVolumen_V', PuntosCotaVolumen_V, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('hDescarga', hDescarga, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('caQE', caQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('cbQE', cbQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('ren', ren, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('Pmax_Gen', Pmax_Gen, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('Qmax_Turb', Qmax_Turb, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('fDispo', fDispo, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('tRepHoras', tRepHoras, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALESPU);

  f.wr('filtracion_Ca', filtracion_Ca, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('filtracion_Cb', filtracion_Cb, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('QaMuySeco', QaMuySeco, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);

  f.wr('cotaMV0', cotaMV0, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('cotaMV1', cotaMV1, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('QMV1', QMV1, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);

  f.wrReferencia('central_lagoDescarga', central_lagoDescarga);
  f.wr('centralesAguasArriba', centralesAguasArriba);

  f.wr('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
  f.wr('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
  f.wr('HayRestriccionQTmin', HayRestriccionQTmin);
  f.wr('QTmin', QTmin);
  f.wr('ImponerQminPorPoste', ImponerQminPorPoste);

  f.wr('hObjetivo', hObjetivo);
  f.wr('cva_delta', delta_cva_ParaControlDeCota);
  f.wr('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
  f.wr('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);

  f.wr('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
  f.wr('PotenciaParaArgentina', P_AG);

  f.wr('QE_CEE', QE_CEE);
  f.wr('hVistaMin', hVistaMin);

  f.wr('cotaControlCrecida_inicio', cotaControlCrecida_inicio);
  f.wr('cotaControlCrecida_pleno', cotaControlCrecida_pleno);
  f.wr('saltoMinimoOperativo', saltoMinimoOperativo);
  f.wr('flg_CalcularEvaporacionDelLago', flg_CalcularEvaporacionDelLago);
  f.wr('flg_CalcularFiltracionDelLago', flg_CalcularFiltracionDelLago);
  f.wr('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
  f.wr('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);

end;

function TFichaHidroConEmbalseBinacional.InfoAd_: string;
begin
  Result := 'PMáxGen= ' + FloatToStrF(Pmax_Gen, ffGeneral, 10, 1) +
    ' MW, ' + 'QMáxTurb= ' + FloatToStrF(Qmax_Turb, ffGeneral, 10, 1) +
    ' Hm^3/s, ' + 'ren= ' + FloatToStrF(ren, ffGeneral, 10, 2) +
    ' p.u., ' + 'fDisp= ' + FloatToStrF(fDispo, ffGeneral, 10, 2) +
    ' p.u., ' + 'tRep= ' + FloatToStrF(tRepHoras, ffGeneral, 10, 1) +
    'h, ' + 'hMín= ' + FloatToStrF(hmin, ffGeneral, 10, 2) + ' m, ' +
    'hMáx= ' + FloatToStrF(hmax, ffGeneral, 10, 2) + ' m, ' +
    'hDescarga= ' + FloatToStrF(hDescarga, ffGeneral, 10, 2) + 'm, ' +
    'Vmáx= ' + FloatToStrF(PuntosCotaVolumen_V[high(PuntosCotaVolumen_V)],
    ffGeneral, 10, 1) + ' m^3' + 'QE_CEE= ' + FloatToStrF(QE_CEE, ffGeneral, 10, 1) +
    ' Hm^3/s, ' + 'hVistaMin= ' + FloatToStrF(hVistaMin, ffGeneral, 10, 2) + ' MWh';
end;

constructor TFichaHidroConEmbalseBinacional.Create_ReadFromText(f: TArchiTexto);
  //var
  //  genOtroPaisSemanal: TDAofNReal;
  //  Vmax, Vhmed: NReal;
begin
  inherited Create_ReadFromText(f);

  PagoPorDisponibilidad_USD_MWh:= 0.0;
  PagoPorEnergia_USD_MWh:= 0.0;

  f.IniciarLecturaRetrasada;
  f.rd('hmin', hmin);
  f.rd('hmax', hmax);
  f.rd('PuntosCotaVolumen_h', PuntosCotaVolumen_h);
  f.rd('PuntosCotaVolumen_V', PuntosCotaVolumen_V);
  f.rd('hDescarga', hDescarga);
  f.rd('caQE', caQE);
  f.rd('cbQE', cbQE);
  f.rd('ren', ren);
  f.rd('Pmax_Gen', Pmax_Gen);
  f.rd('Qmax_Turb', Qmax_Turb);
  f.rd('fDispo', fDispo);
  f.rd('tRepHoras', tRepHoras);
  f.rd('filtracion_Ca', filtracion_Ca);
  f.rd('filtracion_Cb', filtracion_Cb);
  f.rd('QaMuySeco', QaMuySeco);
  f.rd('cotaMV0', cotaMV0);
  f.rd('cotaMV1', cotaMV1);
  f.rd('QMV1', QMV1);
  f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
  f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
  f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
  f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
  f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
  f.rd('QTmin', QTmin);
  f.rd('ImponerQminPorPoste', ImponerQminPorPoste);
  f.rd('hObjetivo', hObjetivo);
  f.rd('cva_delta', delta_cva_ParaControlDeCota);
  f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
  f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
  f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
  f.rd('PotenciaParaArgentina', P_AG);
  f.rd('QE_CEE', QE_CEE);
  f.rd('hVistaMin', hVistaMin);
  f.rd('cotaControlCrecida_inicio', cotaControlCrecida_inicio);
  f.rd('cotaControlCrecida_pleno', cotaControlCrecida_pleno);
  f.rd('saltoMinimoOperativo', saltoMinimoOperativo);

  if f.Version < 83 then
  begin
    flg_CalcularEvaporacionDelLago := True;
    flg_CalcularFiltracionDelLago := True;
  end
  else
  begin
    f.rd('flg_CalcularEvaporacionDelLago', flg_CalcularEvaporacionDelLago);
    f.rd('flg_CalcularFiltracionDelLago', flg_CalcularFiltracionDelLago);
  end;

  if f.Version >= 98 then
  begin
   f.rd('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
   f.rd('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);
  end;

  f.EjecutarLectura;
end;

procedure TFichaHidroConEmbalseBinacional.Free;
begin
  SetLength(PuntosCotaVolumen_h, 0);
  SetLength(PuntosCotaVolumen_V, 0);
  centralesAguasArriba.Free;
  SetLength(P_AG, 0);
  inherited Free;
end;

procedure THidroConEmbalseBinacional_cambioFichaPD(Actor: TCosa);
begin
  (Actor as THidroConEmbalseBinacional).cambioFichaPD;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(THidroConEmbalseBinacional.ClassName,
    THidroConEmbalseBinacional);
  registrarClaseDeCosa(TFichaHidroConEmbalseBinacional.ClassName,
    TFichaHidroConEmbalseBinacional);
end;

procedure AlFinal;
begin
end;

end.
