unit udemandaCombustibleAnioBaseEIndices;

{$MODE Delphi}

interface

uses
  SysUtils,
  uauxiliares,
  xmatdefs,
  math,
  uglobs, uDemandaCombustible, unodos, usimplex,
  uNodoCombustible,
  ufechas,
  ucosa, udatoshorariosdetallados,
  uconstantesSimSEE, uFuentesAleatorias, uFichasLPD;

resourcestring
  rsDemandaGenererada =
    'Demanda Combustible, año base y consumos anuales';

type
  TDemandaCombustibleAnioBaseEIndices = class(TDemandaCombustible)
  public
    ArchiDatos: TArchiRef;
    fIniAnioBase: TFecha;

    Q_demAnioBase: TDAofNReal; // La demanda hora a hora para un año
    // base de 366 días
    invSDem365, invSDem366: NReal; // 1 / la sumatoria de las demandas horarias
    // en [Mm3] del año base considerando 365 y
    // 366 días

    anioIni, anioFin: integer;
    // Primer y último año para el que se especificaron valores
    Mm3_anios: TDAofNReal; // la cantidad de Mm3 demandada para cada año
    // siguiente al de inicio de la simulación

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; nodocomb: TNodoCombustible; falla_profundidad, falla_costo: TDAOfNReal;
      ArchiDatos: string; fuente: TFuenteAleatoria; nombreBorne: string;
      SumarCaudalHr: boolean; xAnioIni, xAnioFin: integer; xMm3_anios: TDAofNReal;
      icf_Fuente: TFuenteAleatoria; icf_NombreBorne: string);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;

    procedure PrepararMemoria(globs: TGlobs); override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    function CaudalFirme: NReal;

    //      procedure prepararSim; override;
    procedure prepararPaso_as; override;
    procedure Free; override;
  private
    CacheActivo: boolean;
    procedure InitFromFile;
  end;

procedure AlInicio;
procedure AlFinal;

implementation

//------------------------------------
// Métodos de TDemandaCombustibleAnioBaseEIndices
//====================================


constructor TDemandaCombustibleAnioBaseEIndices.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; nodocomb: TNodoCombustible;
  falla_profundidad, falla_costo: TDAOfNReal; ArchiDatos: string;
  fuente: TFuenteAleatoria; nombreBorne: string; SumarCaudalHr: boolean;
  xAnioIni, xAnioFin: integer; xMm3_anios: TDAofNReal; icf_Fuente: TFuenteAleatoria;
  icf_NombreBorne: string);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodocomb,
    falla_profundidad, falla_costo, fuente, nombreBorne, icf_Fuente, icf_NombreBorne, SumarCaudalHr);

  if FileExists(ArchiDatos) { *Converted from FileExists*  } then
    self.ArchiDatos := TArchiRef.Create(ArchiDatos)
  else
    raise ExceptionFileNotFound.Create(ArchiDatos, 'Creando el actor ' + ClaseNombre);
  AnioIni := xAnioIni;
  AnioFin := xAnioFin;
  Mm3_anios := xMm3_anios;
  CacheActivo := False;
  InitFromFile;
end;

constructor TDemandaCombustibleAnioBaseEIndices.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.rdArchRef('ArchiDatos', archiDatos);
  f.rd('AnioIni', AnioIni);
  f.rd('AnioFin', AnioFin);
  f.rd('xMm3_anios', Mm3_anios);
  f.EjecutarLectura;
  CacheActivo := False;
  if not ArchiDatos.testear then
    raise ExceptionFileNotFound.Create(ArchiDatos.archi, 'Creando el actor ' +
      ClaseNombre);
  InitFromFile;
end;

procedure TDemandaCombustibleAnioBaseEIndices.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrArchRef('ArchiDatos', ArchiDatos);
  f.wr('AnioIni', AnioIni);
  f.wr('AnioFin', AnioFin);
  f.wr('xMm3_anios', Mm3_anios);
end;

procedure TDemandaCombustibleAnioBaseEIndices.prepararMemoria(globs: TGlobs);

begin
  inherited prepararMemoria(globs);

  InitFromFile;

  if AnioIni < 0 then
  begin
    // esta arreglando el entuerto de la versión anterior a la 12 que no
    // tenía AnioIni y AnioFin

    AnioIni := globs.fechaIniOpt.anio;
    Aniofin := globs.fechaFinOpt.anio;
    if length(Mm3_anios) < (AnioFin - AnioIni + 1) then
      AnioFin := AnioIni + high(Mm3_anios);
    globs.Alerta('Ojo, revise el actor Demanda: ' + nombre +
      '. Fue transformado desde una versión anterior que no especificaba el año de inicio y de fin.');
  end;

  if (anioIni > globs.fechaIniSim.anio) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el inicio de la Simulación, se usará el primer año para cubrir el hueco.');

  if (anioIni > globs.fechaIniOpt.anio) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el inicio de la Optimización, se usará el primer año para cubrir el hueco.');

  if (globs.fechaFinSim.anio > anioFin) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el final de la Simulación, se usará el último año para cubrir el hueco.');

  if (globs.fechaFinOpt.anio > anioFin) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el final de la Optimización, se usará el último año para cubrir el hueco.');

end;

function TDemandaCombustibleAnioBaseEIndices.infoAd_: string;
begin
  Result := ArchiDatos.archi;
end;

class function TDemandaCombustibleAnioBaseEIndices.DescClase: string;
begin
  Result := rsDemandaGenererada;
end;

function TDemandaCombustibleAnioBaseEIndices.CaudalFirme: NReal;
begin
  Result := paUnidades.nUnidades[0] * vmax(self.QHoraria);
end;

procedure TDemandaCombustibleAnioBaseEIndices.prepararPaso_as;
var
  factCrecimiento: NReal;
  kAnio: integer; // indice del año del paso dentro del vector de años especificados
  jHoraDelPaso: integer; // hora desde el inicio delpaso 0 ... durpaso-1
  jHoraDelAnio_ultimaHoraDelAnio: integer;
  // auxiliar, se calcula al actualizar el factor
  jHoraDelAnio_desde, jHoraDelAnio_hasta: integer; // limites del paso
  jLimite: integer;

  dtInicioDelAnio: TDateTime;
  rh: NReal;
  ih: integer;
  alfa: NReal;
  Q1, Q2: NReal;

  procedure ActualizarFactorYUltimaHoraDelAnio(anio: integer);
  begin
    kAnio := anio - AnioIni;

    // si actua esta extensión ya se lo avisamos al preparar memoria.
    if kAnio < 0 then
      kAnio := 0
    else if kAnio > high(Mm3_anios) then
      kAnio := high(Mm3_anios);

    if not IsLeapYear(anio) then
    begin
      factCrecimiento := Mm3_anios[kAnio] * invSDem365;
      jHoraDelAnio_UltimaHoraDelAnio := 365 * 24 - 1;
    end
    else
    begin
      factCrecimiento := Mm3_anios[kAnio] * invSDem366;
      jHoraDelAnio_UltimaHoraDelAnio := 366 * 24 - 1;
    end;
  end;

begin
  ActualizarFactorYUltimahoraDelAnio(globs.AnioInicioDelPaso);


  if globs.SalaMinutal then
  begin
    rh:= globs.FechaInicioDelpaso.diasDesdeElInicioDelAnio * 24.0;
    ih:= trunc( rh );
    alfa:= frac( rh );
    Q1:=  Q_demAnioBase[ ih ];
    inc( ih );
    if ih <= high( Q_demAnioBase ) then
      Q2:= Q_demAnioBase[ ih ]
    else
      Q2:= Q_demAnioBase[ 0 ];
    QHoraria[0] := (Q1 *(1-alfa)+Q2*alfa) * factCrecimiento;
  end
  else
  begin
    // inicializamos puntero de lectura sobre el anio base.
    jHoraDelAnio_desde := globs.FechaInicioDelpaso.horasDesdeElInicioDelAnio;
    jHoraDelAnio_hasta := ceil( jHoraDelAnio_desde + globs.HorasDelPaso - 1 );

  //Cargamos los datos que vamos a usar en el paso
  if jHoraDelAnio_hasta <= jHoraDelAnio_UltimaHoraDelAnio then
  begin // el paso está íntegro dentro del mismo año no necesita partir el for
    for jHoraDelPaso := 0 to high( QHoraria ) do
      QHoraria[jHoraDelPaso] :=
        Q_demAnioBase[jHoraDelAnio_desde + jHoraDelPaso] * factCrecimiento;
  end
  else // globs.FechaInicioDelpaso.anio < globs.FechaFinDelPaso.anio
  begin // el paso comienza en un año y termina en el siguiente
    jLimite := (jHoraDelAnio_UltimaHoraDelAnio - jHoraDelAnio_desde);
    for jHoraDelPaso := 0 to jLimite do
      QHoraria[jHoraDelPaso] :=
        Q_demAnioBase[jHoraDelAnio_desde + jHoraDelPaso] * factCrecimiento;

    ActualizarFactorYUltimaHoraDelAnio(globs.AnioInicioDelPaso + 1);

    for jHoraDelPaso := jLimite +1 to high( QHoraria ) do
      QHoraria[jHoraDelPaso] := Q_demAnioBase[jHoraDelPaso] * factCrecimiento;
  end;
  end;

  inherited prepararPaso_as;
end;

procedure TDemandaCombustibleAnioBaseEIndices.Free;
begin
  fIniAnioBase.Free;
  if Q_demAnioBase <> nil then
    SetLength(Q_demAnioBase, 0);
  SetLength(Mm3_anios, 0);
  ArchiDatos.Free;
  inherited Free;
end;

procedure TDemandaCombustibleAnioBaseEIndices.InitFromFile;
var
  datos: TDatosHorariosDetallados;
  i: integer;
begin
  datos := TDatosHorariosDetallados.Create(self.ArchiDatos.archi);
  SetLength(Q_demAnioBase, 366 * 24);
  datos.ReadBuff_horario(Q_demAnioBase, datos.fechaPrimerDia);

  fIniAnioBase := TFecha.Create_Dt(datos.fechaPrimerDia);

  invSDem365 := 0;
  for i := 0 to 365 * 24 - 1 do
  begin
    invSDem365 := invSDem365 + Q_demAnioBase[i];
  end;
  invSDem366 := invSDem365;
  for i := 365 * 24 to 366 * 24 - 1 do
    invSDem366 := invSDem366 + Q_demAnioBase[i];

  invSDem365 := 1.0E6 / invSDem365;
  invSDem366 := 1.0E6 / invSDem366;

  for i:= 0 to high( Q_demAnioBase ) do
    Q_demAnioBase[i]:= Q_demAnioBase[i] / 3600.0; // pasamos de m3/h a m3/s

  datos.Free;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TDemandaCombustibleAnioBaseEIndices.ClassName, TDemandaCombustibleAnioBaseEIndices);
end;

procedure AlFinal;
begin
end;

end.
