unit uRobotHttpPost;

(*
Proyecto: SimSEE
Autor: Ruben Chaer
Fecha: 2014-03-29
Descripción:  esta unidad permite crear agentes para realizar
  consultas a un sitio web mediante envio de formularios POST.
  Se implementó para la consulta de pronósticos en SimSEE.

*)
{$mode delphi}

interface

uses
  Classes, httpsend, SysUtils;

type
(***
   La idea es que se crea el Robot dando una url a la que se envian
   las consultas.
   Los resultados de las consultas se devuelven en un TStringList
   Para facilitar el FILTRADO de resultados, al inicializar el
   Robot además de la url se puden pasar las string InicioRes y FinRes
   si esas sring son diferentes de vacio, se elminan toas las lineas
   antes de InicioRes (inclusive) y todas las líneas posteriores a FinRes (inclusive)

   Para definir los parámetros del formulario se utilizan los métodos
   ClearCampos que elimina todos los campos que estuviesen definidos
   addCampo permite definir un nuevo campo en el formulario pasando
      nombre y valor. Esta función retorna el ordinal identificador
      del campo agregado en el formulario.
   setCampo permite fijar un nuevo valor al campo cuyo ordinal es el
      entero kPar pasado como parámetro.



***)

  TRobotHttpPost = class
  private
    url: string; { ej: http://simsee.org/pronosticos/index.php }
    InicioRes: string; {ej: '+inicio'}
    FinRes: string; {ej: '+fin' }

    proxy_host, proxy_port, proxy_user, proxy_pass: string;

    campos_nombres: TStringList;
    campos_valores: TStringList;

  public
    constructor Create(url, InicioRes, FinRes: string);
    procedure set_proxy(host, port, user, pass: string);
    procedure ClearCampos;
    function AddCampo(nombre, valor: string): integer;
    procedure SetCampo(kPar: integer; NuevoValor: string);
    function post( Metodo: String ): TStringList;

  private
    function ProxyHttpPostURL(
     const URL, URLData: string; const Data: TStream; Metodo: string ): boolean;
  end;

implementation

function URLEncode(Str: string): string;
var
  i, j: integer;
  res, s2: string;
begin
  //  result:= synacode.EncodeURL( str );
  setlength(res, length(str) * 3);
  j := 1;
  for i := 1 to Length(Str) do
    if Str[i] in ['A'..'Z', 'a'..'z', '0'..'9', '-', '_', '.'] then
    begin
      Res[j] := Str[i];
      Inc(j);
    end
    else
    begin
      Res[j] := '%';
      Inc(j);
      s2 := IntToHex(Ord(Str[i]), 2);
      Res[j] := s2[1];
      Inc(j);
      Res[j] := s2[2];
      Inc(j);
    end;
  setlength(res, j - 1);
  Result := res;
end;


constructor TRobotHttpPost.Create(url, InicioRes, FinRes: string);
begin
  inherited Create;
  self.url := url;
  self.inicioRes := inicioRes;
  self.finRes := finRes;

  campos_nombres := TStringList.Create;
  campos_valores := TStringList.Create;
  proxy_host := '';
  proxy_port := '';
  proxy_user := '';
  proxy_host := '';
end;


procedure TRobotHttpPost.set_proxy(host, port, user, pass: string);
begin
  proxy_host := host;
  proxy_port := port;
  proxy_user := user;
  proxy_host := pass;
end;

procedure TRobotHttpPost.ClearCampos;
begin
  campos_nombres.Clear;
  campos_valores.Clear;
end;

function TRobotHttpPost.AddCampo(nombre, valor: string): integer;
begin
  campos_nombres.Add(nombre);
  campos_valores.add(valor);
  Result := campos_nombres.Count - 1;
end;

procedure TRobotHttpPost.SetCampo(kPar: integer; NuevoValor: string);
begin
  campos_valores[kPar] := NuevoValor;
end;


function TRobotHttpPost.ProxyHttpPostURL(const URL, URLData: string;
  const Data: TStream; Metodo: string ): boolean;
var
  HTTP: THTTPSend;
begin
  HTTP := THTTPSend.Create;
  try
    HTTP.ProxyHost := proxy_host;
    HTTP.ProxyPort := proxy_port;
    HTTP.ProxyUser := proxy_user;
    HTTP.ProxyPass := proxy_pass;
    HTTP.Document.Write(Pointer(URLData)^, Length(URLData));
    HTTP.MimeType := 'application/x-www-form-urlencoded';
//    HTTP.UserAgent:= 'Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.137 Safari/537.36';
//Result := HTTP.HTTPMethod('POST', URL);
    Result := HTTP.HTTPMethod( Metodo, URL);
    Data.CopyFrom(HTTP.Document, 0);
  finally
    HTTP.Free;
  end;
end;

function TRobotHttpPost.post( Metodo: string ): TStringList;
var
  rb: string;
  Content_Length: cardinal;
  buff: string;
  ipStr: string;
  error: string;
  res: TStringList;
  k, j: integer;
  st: TMemoryStream;
  buscando: boolean;
  kPar: integer;
  rs: string;
begin
  rb := '';
  for kPar := 0 to campos_nombres.Count - 1 do
  begin
    if kPar > 0 then
      rb := rb + '&';
    rb := rb + URLEncode(campos_nombres[kPar]) + '=' + URLEncode(campos_valores[kPar]);
  end;


  st := TMemoryStream.Create;
  try
    ProxyHTTPpostURL(url, rb, st, Metodo);
    st.Seek(0, soFromBeginning);
    res := TStringList.Create;
    res.LoadFromStream(st);
  finally
    st.Free;
  end;


  if InicioRes <> '' then
  begin

    // Borro las líneas anteriores al inicio
    buscando := True;
    while (res.Count > 0) and buscando do
    begin
      rs:= res[0];
      writeln( rs );
      if pos(InicioRes, rs) = 1 then
        buscando := False
      else
        res.Delete(0);
    end;

    if buscando then
    begin
      res.Free;
      raise Exception.Create(
        'TRobotHttpPost.post: error, no se recibió INICIO ' + InicioRes);
    end;
  end;

  if FinRes <> '' then
  begin
    buscando := True;
    while (res.Count > 0) and buscando do
    begin
      rs:= res[res.Count - 1];
      writeln( rs );
      if pos(FinRes, rs) = 1 then
        buscando := False
      else
        res.Delete(res.Count - 1);
    end;

    if buscando then
    begin
      res.Free;
      raise Exception.Create(
        'TRobotHttpPost.post: error, no se recibió FINAL ' + FinRes);
    end;

  end;

  Result := res;
end;




end.
