(* Oct.2005, rch@todo.com.uy *)
unit uregresion;

interface

uses
  SysUtils, Classes, xmatdefs, matreal, Math;

(******
retorna un filtro recursivo que explica el vector y con la mnima
varianza de la entrada
yk = sum( bj y(k-j) ; j= 1..orden ) + uk

siendo bj el elemento j del resultado

si filtrar = true, solo se tienen en cuenta las muestras con valor
mayor que umbralFiltro. Esto es til, para filtrar las muestras dudosas
para lo cual se les pone un valor negativo, menor que cualquier valor
que pueda tomar el proceso real y se fija el filtro para que no considere
esas muestras.

*******)
function calcmincuad_rh(y: TVectR; orden: integer; filtrar: boolean;
  umbralFiltro: NReal): TVectR;

(***********
Clculo de las matrices MTM y MTY del sistema MTM . A = MTY para
el racimo de vectores (series) pasado como parmetros.
El el orden del filtro es la cantidad de pasos del pasado considerado
para cada serie. La matriz A es de [orden*nseries x orden*nseries]

si filtrar = true, solo se tienen en cuenta las muestras con valor
mayor que umbralFiltro. Esto es til, para filtrar las muestras dudosas
para lo cual se les pone un valor negativo, menor que cualquier valor
que pueda tomar el proceso real y se fija el filtro para que no considere
esas muestras.

***********)
procedure calc_sistema(var XXt, XYt, YYt: TMatR; series: array of TVectR;
  orden: integer; filtrar: boolean; umbralFiltro: NReal; Normalizar: boolean);


(*****
retorna un filtro recursivo multi-variable que explica los vectores
"series" como la combinacin lineal de sus pasados (de 1 a orden) ms un
vector de mnima varianza

si filtrar = true, solo se tienen en cuenta las muestras con valor
mayor que umbralFiltro. Esto es til, para filtrar las muestras dudosas
para lo cual se les pone un valor negativo, menor que cualquier valor
que pueda tomar el proceso real y se fija el filtro para que no considere
esas muestras.

Retorna la Matriz A' del filtro lineal que explica
X[k+1] = A [X[k]; X[k-1] ; ...;  X[k- (orden-1)] + B R

Y = X[k+1]
X = [X[k]; X[k-1] ; ...;  X[k- (orden-1)]

Y = A X + B R


Donde R / < R * R' > = I


***)
function calcmincuad_mrh(series: array of TVectR; orden: integer;
  filtrar: boolean; umbralFiltro: NReal; var BBt: TMatR): TMatR;



(*****
retorna un filtro recursivo multi-ciclo y  multi-variable que explica los vectores
my como la combinacin lineal de sus pasados (de 1 a orden) ms un
vector de mnima varianza.

si filtrar = true, solo se tienen en cuenta las muestras con valor
mayor que umbralFiltro. Esto es til, para filtrar las muestras dudosas
para lo cual se les pone un valor negativo, menor que cualquier valor
que pueda tomar el proceso real y se fija el filtro para que no considere
esas muestras.

El parmetro overlapping indica en cuantos pasos anteriores y posteriores
debe ser considerado una muestra en la determinacin del filtro. Por ejemplo
si overlapping=3 una muestra ser considerada en el paso de tiempo que corresponde
y en los tres pasos anteriores y los tres posteriores.

El resultado es dos array de las nPasosPorCiclo matrices del filtro.
Una matriz A y una B para cada paso de ciclo.

El resultado es la cantidad de FALLOS en la resolucin del problema de mnimos
cuadrados. Si todo anduvo OK, es CERO. Para los sitemas que no pudo resolver,
 mcA[k] y mcB[k] son NIL.

Si parmetro Normalizar = TRUE , se supone que las series son de Norma 1 y
como el tratamiento multi-ciclo puede romper con esa caracterstica se inenta
volver a recomponer sobre los coeficientes de correlacin con que se calculan
las matrices de aproximacin por mnimos cuadrados.

El parmetro series_Nombres sirve para etiquetar la impresin a archivo
de la matriz de covarianzas. Poner NIL si no corresponde
***)
function calcmincuad_mrh_mc(series_X: array of TVectR; orden: integer;
  filtrar: boolean; umbralFiltro: NReal; nPasosPorCiclo: integer;
  overlap: integer; var mcA, mcB: TDAOfMatR; Normalizar: boolean;
  series_Nombres: TStringList;
  var VarNoExplicado: NReal  // variaza de lo no explicado promedio
  ): integer;




implementation



function calcmincuad_rh(y: TVectR; orden: integer; filtrar: boolean;
  umbralFiltro: NReal): TVectR;

var
  m, i: TMatR;
  k, j: integer;
  a: NReal;
  invertible: boolean;
  e10: integer;

begin
  m := TMatR.Create_Init(orden, orden);
  i := TMatR.Create_Init(orden, 1);

  for k := 1 to orden do
  begin
    for j := k to orden do
    begin
      a := y.coefcorr(y, j - k, filtrar, umbralFiltro);
      m.pon_e(k, j, a);
      if k <> j then
        m.pon_e(j, k, a);
    end;
    a := y.coefcorr(y, k, filtrar, umbralFiltro);
    i.pon_e(k, 1, a);
  end;

  m.Escaler(i, invertible, e10);
  Result := i.crear_columna(1);
end;


(***********
Clculo de las matrices MTM y MTY del sistema MTM . A = MTY para
el racimo de vectores (series) pasado como parmetros.
El el orden del filtro es la cantidad de pasos del pasado considerado
para cada serie.si p es el nmero de series,
dim= orden* p La matriz MTM es de [dim x dim ]
Las matrices A  y MTY son de [dim x p]
***********)

(*
procedure calc_sistema(var MTM: TMatR; var MTY: TMatR; series: array of TVectR;
  orden: integer; filtrar: boolean; umbralFiltro: NReal);
var
  k, j: integer;
  p:    integer;
  ks, kcof: integer;
  js, jcof: integer;
  dim:  integer;
  jj:   integer;
begin

  p   := length(series);
  dim := p * orden;

  MTM := TMatR.Create_Init(dim, dim);
  MTY := TMatR.Create_Init(dim, p);

  k := 0; // inicializo el ndice de fila

  for kcof := 1 to orden do
    for ks := 1 to p do
    begin

      k := k + 1; // ndice de fila
      // clculo del coeficiente de MTM
      j := 0; //inicializo el ndice de columna

      for jcof := 1 to orden do
        for js := 1 to p do

        begin
          j := j + 1; // incremento el ndice de columna de MTM
          MTM.pon_e(k, j, series[ks - 1].coefcorr(series[js - 1],
            kcof - jcof, filtrar, umbralFiltro));
        end;

      // ahora calculamos el coeficiente de MTY
      for jj := 1 to p do
        MTY.pon_e(k, jj, series[ks - 1].coefcorr(series[jj - 1], kcof,
          filtrar, umbralFiltro));

    end;
end;
*)

procedure calc_sistema(var XXt, XYt, YYt: TMatR; series: array of TVectR;
  orden: integer; filtrar: boolean; umbralFiltro: NReal; Normalizar: boolean);

var
  Y, X: TVectR; // una muestra
  NDatos: integer; // cantidad de puntos donde es posible calcular el filtro.
  NSeries: integer;
  dim: integer;
  cnt_muestras, cnt_filtrados: integer;
  kMuestra: integer;
  kFil, jCol: integer;


  function LlenarMuestra(kMuestra: integer): boolean;
  var
    jserie: integer;
    kretardo: integer;
    m: NReal;
    res: boolean;
  begin
    res := True;
    for kretardo := 1 to orden do
      for jserie := 1 to NSeries do
      begin
        m := series[jserie - 1].pv[kMuestra + orden - kretardo];
        if filtrar then
          res := res and (m > umbralFiltro);
        X.pon_e((kretardo - 1) * NSeries + jserie, m);
      end;

    for jserie := 1 to NSeries do
    begin
      m := series[jserie - 1].pv[kMuestra + orden];
      if filtrar then
        res := res and (m > umbralFiltro);
      Y.pon_e(jserie, m);
    end;
    Result := res;
  end;

begin
  NSeries := length(series);
  NDatos := series[0].n - orden; // largo de las series de datos
  dim := NSeries * orden;
  X := TVectR.Create_init(dim);
  Y := TVectR.Create_init(NSeries);

  XXt := TMatR.Create_Init(dim, dim);
  XYt := TMatR.Create_Init(dim, NSeries);
  YYt := TMatR.Create_Init(NSeries, NSeries);

  cnt_muestras := 0;
  cnt_filtrados := 0;

  for kMuestra := 1 to NDatos do
  begin
    if LlenarMuestra(kMuestra) then
    begin
      Inc(cnt_muestras);
      for kFil := 1 to Dim do
      begin
        for jCol := kFil to Dim do
          XXt.acum_e(kFil, jCol, X.e(kFil) * X.e(jCol));
        for jCol := 1 to NSeries do
          XYt.acum_e(kFil, jCol, X.e(kFil) * Y.e(jCol));
      end;

      // llenamos YYt
      for kFil := 1 to NSeries do
        for jCol := kFil to NSeries do
          YYt.acum_e(kFil, jCol, Y.e(kFil) * Y.e(jCol));
    end
    else
      Inc(cnt_filtrados);
  end;

  // promedios
  for kFil := 1 to Dim do
  begin
    for jCol := kFil to Dim do
      XXt.pon_e(kFil, jCol, XXt.e(kFil, jCol) / cnt_Muestras);
    for jCol := 1 to NSeries do
      XYt.pon_e(kFil, jCol, XYt.e(kFil, jCol) / cnt_Muestras);
  end;
  for kFil := 1 to NSeries do
    for jCol := kFil to NSeries do
      YYt.pon_e(kFil, jCOl, YYt.e(kFil, jCol) / cnt_Muestras);

  if Normalizar then
  begin
    // bueno si est marcado NORMALIZAR, las diagonales de XXt e YYt tienen que
    // ser 1s (unos )
    // Calculo en X e Y las respectivas desviaciones estndar.
    for kFil := 1 to Dim do
      X.pon_e(kFil, sqrt(XXt.e(kFil, kFil)));
    for kFil := 1 to NSeries do
      Y.pon_e(kFil, sqrt(YYt.e(kFil, kFil)));

    // ahora dividimos por las desviaciones estndar.
    for kFil := 1 to Dim do
    begin
      for jCol := kFil to Dim do
        XXt.pon_e(kFil, jCol, XXt.e(kFil, jCol) / (X.e(kFil) * X.e(jCol)));
      for jCol := 1 to NSeries do
        XYt.pon_e(kFil, jCol, XYt.e(kFil, jCol) / (X.e(kFil) * Y.e(jCOl)));
    end;

    for kFil := 1 to NSeries do
      for jCol := kFil to NSeries do
        YYt.pon_e(kFil, jCol, YYt.e(kFil, jCol) / (Y.e(kFil) * Y.e(jCol)));
  end;

  // completamos los tringulos inferiores de XXt e YYt
  for kFil := 2 to Dim do
    for jCol := 1 to kFil - 1 do
      XXt.pon_e(kFil, jCol, XXt.e(jCol, kFil));

  for kFil := 2 to NSeries do
    for jCol := 1 to kFil - 1 do
      YYt.pon_e(kFil, jCol, YYt.e(jCol, kFil));


  Y.Free;
  X.Free;

end;




function calcmincuad_mrh(series: array of TVectR; orden: integer;
  filtrar: boolean; umbralFiltro: NReal; var BBt: TMatR): TMatR;
var
  XXt, XYt, YYt, YXt: TMatR;
  invertible: boolean;
  k, j, h, jRet: integer;
  e10: integer;

begin
  calc_sistema(XXt, XYt, YYt, series, orden, filtrar, umbralFiltro, True);

  XXt.WriteArchiXLT('Matriz_XXt.xlt');

  YXt := XYt.crear_transpuesta;

  XXt.Escaler(XYt, invertible, e10); // At
  if not invertible then
  begin
    raise Exception.Create('CalcMinCuad ... error. el sistema no es invertible.');
  end;
  BBt := TMatR.Create_Init(length(series), length(series));
  // Y = AX + B R
  // YYt = A XXt At + BBt => BBt = YYt - A XXt At
  // YXt = A XXt => BBt = YYt - YXt * At

  BBt.Mult(YXt, XYt);
  BBt.PorReal(-1);
  BBt.sum(YYt);

  YYt.Free;
  YXt.Free;
  XXt.Free;
  Result := XYt;
end;




(***********
Clculo de las matrices MTM y MTY del sistema MTM . A = MTY para
el paso kPaso del racimo de vectores (series) pasado como parmetros,
considerando que dichas series forman ciclos de nPasosPorCiclo.
kPaso puede tomar valores en [0..nPasosPorCiclo-1]

Para la formacin de las matrices se considera el permetro overlapping

El el orden del filtro es la cantidad de pasos del pasado considerado
para cada serie.
si p es el nmero de series,
dim= orden* p
La matriz MTM es de [dim x dim ]
Las matrices A  y MTY son de [dim x p]

Y = A X * B R

YYt es la matriz
 MTM = XXt

Si el parmetro Normalizar es true se est indicando que las series
se suponen ergodizadas y de norma 1 y como el Multi-Ciclo puede hacer
que se "rompa la normalidad" se debe recomponer sobre las matrices de clculo.
***********)

(*
procedure calc_sistema_MultiCiclo(
  var MTM: TMatR; var MTY: TMatR;
  var YYt: TMatR;
  series: array of TVectR; orden: integer; filtrar: boolean;
  umbralFiltro: NReal;
  kPaso: integer; nPasosPorCiclo: integer;
  overlapping: integer; Normalizar: boolean);
var
  k, j: integer;
  NSeries:    integer;
  k_serie, k_retardo: integer;
  j_serie, j_retardo: integer;
  dim:  integer;
  jj:   integer;
  cnt_muestras: integer;
begin

  NSeries  := length(series);
  dim := NSeries * orden;

  MTM := TMatR.Create_Init(dim, dim);
  MTY := TMatR.Create_Init(dim, NSeries );
  YYt := TMatR.Create_Init( NSeries, NSeries );

  k := 0; // inicializo el ndice de fila
  for k_retardo := 1 to orden do
  begin
    for k_serie := 1 to NSeries do
    begin
      k := k + 1; // ndice de fila

      // clculo del coeficiente de MTM
      j := 0; //inicializo el ndice de columna
      for j_retardo := 1 to orden do
        for j_serie := 1 to NSeries do
        begin
          j := j + 1; // incremento el ndice de columna de MTM
          MTM.pon_e(k, j,
            series[k_serie - 1].coefcorr_MultiCiclo(
            series[j_serie - 1],
            k_retardo - j_retardo,
            filtrar, umbralFiltro,
            kPaso, nPasosPorCiclo, overlapping, cnt_muestras, Normalizar,
            'mtm_x'+IntToStr(k_Serie)+'_x'+IntToStr(j_serie)+'_'+IntToSTr( k_retardo-j_retardo)+'_'+IntToStr( kPaso )+'.xlt' ));
          if cnt_muestras = 0 then
            raise Exception.Create(
              'OJO, uregresion.calc_sistema_MultiCiclo; MTM_ cnt_muestras = 0 ');
        end;

      // ahora calculamos el coeficiente de MTY
      for jj := 1 to NSeries do
        MTY.pon_e(k, jj,
          series[k_serie - 1].coefcorr_MultiCiclo(series[jj - 1], k_retardo,
          filtrar, umbralFiltro, kPaso, nPasosPorCiclo, overlapping,
          cnt_muestras, Normalizar,
           'mty_x'+IntToStr(k_Serie)+'_y'+IntToStr(jj)+'_'+IntToSTr( k_retardo )+'_'+IntToStr( kPaso )+'.xlt' ));


        if cnt_muestras = 0 then
        raise Exception.Create(
          'OJO, uregresion.calc_sistema_MultiCiclo; MTY_ cnt_muestras = 0 ');

    end;
  end;

  // ahora calculamos YYt
  for k:= 1 to NSeries do
    for jj := k to NSeries do
    begin
        YYt.pon_e(k, jj,
          series[k - 1].coefcorr_MultiCiclo(
            series[jj - 1],
            0, // kretardo
            filtrar,
            umbralFiltro,
            kPaso +1 ,  // el paso siguiente.
            nPasosPorCiclo,
            overlapping,
            cnt_muestras,
            Normalizar,
           'mty_x'+IntToStr(k_Serie)+'_y'+IntToStr(jj)+'_'+IntToSTr( k_retardo )+'_'+IntToStr( kPaso )+'.xlt' ));

        if jj > k then
          YYt.pon_e( jj, k, YYt.e( k, jj ) );
    end;

end;

*)


procedure calc_sistema_MultiCiclo(var XXt, XYt, YYt: TMatR;
  series: array of TVectR; orden: integer; filtrar: boolean;
  umbralFiltro: NReal; kPaso: integer; nPasosPorCiclo: integer;
  overlap: integer; Normalizar: boolean);

var
  Y, X: TVectR; // una muestra
  NDatos: integer; // cantidad de puntos donde es posible calcular el filtro.
  NSeries: integer;
  dim: integer;
  cnt_muestras, cnt_filtrados: integer;
  kMuestra, kPos: integer;
  kFil, jCol: integer;


  function LlenarMuestra(kMuestra: integer): boolean;
  var
    jserie: integer;
    kretardo: integer;
    m: NReal;
    res: boolean;
  begin
    res := True;
    for kretardo := 1 to orden do
      for jserie := 1 to NSeries do
      begin
        m := series[jserie - 1].pv[kMuestra + orden - kretardo];
        if filtrar then
          res := res and (m > umbralFiltro);
        X.pon_e((kretardo - 1) * NSeries + jserie, m);
      end;

    for jserie := 1 to NSeries do
    begin
      m := series[jserie - 1].pv[kMuestra + orden];
      if filtrar then
        res := res and (m > umbralFiltro);
      Y.pon_e(jserie, m);
    end;

    Result := res;
  end;

begin
  NSeries := length(series);
  NDatos := series[0].n - orden; // largo de las series de datos
  dim := NSeries * orden;
  X := TVectR.Create_init(dim);
  Y := TVectR.Create_init(NSeries);


  XXt := TMatR.Create_Init(dim, dim);
  XYt := TMatR.Create_Init(dim, NSeries);
  YYt := TMatR.Create_Init(NSeries, NSeries);

  cnt_muestras := 0;
  cnt_filtrados := 0;

  for kMuestra := 1 to NDatos do
  begin
    kPos := ((kMuestra - 1) mod NPasosPorCiclo);
    if (kPos >= (kPaso - overlap)) and (kPos <= (kPaso + overlap)) then
      if LlenarMuestra(kMuestra) then
      begin
        Inc(cnt_muestras);
        for kFil := 1 to Dim do
        begin
          for jCol := kFil to Dim do
            XXt.acum_e(kFil, jCol, X.e(kFil) * X.e(jCol));
          for jCol := 1 to NSeries do
            XYt.acum_e(kFil, jCol, X.e(kFil) * Y.e(jCol));
        end;

        // llenamos YYt
        for kFil := 1 to NSeries do
          for jCol := kFil to NSeries do
            YYt.acum_e(kFil, jCol, Y.e(kFil) * Y.e(jCol));
      end
      else
        Inc(cnt_filtrados);
  end;

  // promedios
  for kFil := 1 to Dim do
  begin
    for jCol := kFil to Dim do
      XXt.pon_e(kFil, jCol, XXt.e(kFil, jCol) / cnt_Muestras);
    for jCol := 1 to NSeries do
      XYt.pon_e(kFil, jCol, XYt.e(kFil, jCol) / cnt_Muestras);
  end;
  for kFil := 1 to NSeries do
    for jCol := kFil to NSeries do
      YYt.pon_e(kFil, jCOl, YYt.e(kFil, jCol) / cnt_Muestras);

  if Normalizar then
  begin
    // bueno si est marcado NORMALIZAR, las diagonales de XXt e YYt tienen que
    // ser 1s (unos )
    // Calculo en X e Y las respectivas desviaciones estndar.
    for kFil := 1 to Dim do
      X.pon_e(kFil, sqrt(XXt.e(kFil, kFil)));
    for kFil := 1 to NSeries do
      Y.pon_e(kFil, sqrt(YYt.e(kFil, kFil)));

    // ahora dividimos por las desviaciones estndar.
    for kFil := 1 to Dim do
    begin
      for jCol := kFil to Dim do
        XXt.pon_e(kFil, jCol, XXt.e(kFil, jCol) / (X.e(kFil) * X.e(jCol)));
      for jCol := 1 to NSeries do
        XYt.pon_e(kFil, jCol, XYt.e(kFil, jCol) / (X.e(kFil) * Y.e(jCOl)));
    end;

    for kFil := 1 to NSeries do
      for jCol := kFil to NSeries do
        YYt.pon_e(kFil, jCol, YYt.e(kFil, jCol) / (Y.e(kFil) * Y.e(jCol)));
  end;

  // completamos los tringulos inferiores de XXt e YYt
  for kFil := 2 to Dim do
    for jCol := 1 to kFil - 1 do
      XXt.pon_e(kFil, jCol, XXt.e(jCol, kFil));

  for kFil := 2 to NSeries do
    for jCol := 1 to kFil - 1 do
      YYt.pon_e(kFil, jCol, YYt.e(jCol, kFil));


  Y.Free;
  X.Free;

end;


function calcmincuad_mrh_mc(series_X: array of TVectR; orden: integer;
  filtrar: boolean; umbralFiltro: NReal; nPasosPorCiclo: integer;
  overlap: integer; var mcA, mcB: TDAOfMatR; Normalizar: boolean;
  series_Nombres: TStringList;
  var VarNoExplicado: NReal  // variaza de lo no explicado promedio
  ): integer;

var
  //  mXX0,
  Ti, mBBt, mYYt: TMatR;
  mXXt, mXXst: TMatR;

  kPaso: integer;
  invertible: boolean;
  cnt_fallos: integer;
  k, j: integer;

  sal: TextFile;
  kFil, jCol: integer;
  e10: integer;

begin
  setlength(mcA, nPasosPorCiclo);
  setlength(mcB, nPasosPorCiclo);

  cnt_fallos := 0;
  VarNoExplicado := 0;


  assignfile(sal, 'covarianzas_XX0.xlt');
  rewrite(sal);

  Write(sal, '-');
  if series_Nombres <> nil then
  begin
    for k := 0 to Series_NOmbres.Count - 1 do
      for j := 0 to Series_NOmbres.Count - 1 do
        if j >= k then
          Write(sal, #9, Series_Nombres[k] + '-' + Series_Nombres[j]);
    writeln(sal);
  end;

  for kPaso := 0 to nPasosPorCiclo - 1 do
  begin
    calc_sistema_MultiCiclo(
      mXXt, mXXst, mYYt,
      series_X, orden, filtrar, umbralFiltro,
      kPaso, nPasosPorCiclo, overlap, Normalizar);

    Write(sal, kPaso);
    for kFil := 1 to mXXt.nf do
      for jCol := kFil to mXXt.nc do
        Write(sal, #9, mXXt.e(kFil, jCol));
    writeln(sal);

    mXXt.WriteArchiXLT('mxxt_' + IntToStr(kPaso) + '.xlt');
    mXXst.WriteArchiXLT('mxxst_' + IntToStr(kPaso) + '.xlt');
    mYYt.WriteArchiXLT('myyt_' + IntToStr(kPaso) + '.xlt');


    Ti := mXXst.crear_transpuesta;

    mXXt.Escaler(mXXst, invertible, e10);
    if not invertible then
    begin
      Inc(cnt_fallos);
      mcA[kPaso] := nil;
      mcB[kPaso] := nil;
      mYYt.Free;
      Ti.Free;
      mXXst.Free;
      mXXt.Free;
    end
    else
    begin
      mBBt := TMatR.Create_Init(length(Series_X), length(Series_X));

      mBBt.Mult(Ti, mXXst);
      mBBt.PorReal(-1);
      mBBt.sum(mYYt);

      mXXst.transponer;
      mcA[kPaso] := mXXst;

      mcB[kPaso] := mBBt.raiz_Cholesky;
      if mcB[kPaso] <> nil then
      begin
        VarNoExplicado := VarNoExplicado + mBBt.Traza;
      end
      else
        Inc(cnt_fallos);

      Ti.Free;
      mXXt.Free;
      mBBt.Free;

    end;
  end;
  closefile(sal);

  if cnt_fallos = 0 then
    VarNoExplicado := VarNoExplicado / length(Series_X) / nPasosPorCiclo
  else
    VarNoExplicado := -1;
  Result := cnt_fallos;
end;


end.

















